'''
Language parser for Solidity
'''

from .clike import CCppCommentsMixin
from .code_reader import CodeReader
from .golike import GoLikeStates


class SolidityReader(CodeReader, CCppCommentsMixin):

    ext = ['sol']
    language_names = ['solidity']
    _conditions = set(['if', 'for', 'while', '&&', '||', '?'])

    def __init__(self, context):
        super(SolidityReader, self).__init__(context)
        self.parallel_states = [SolidityStates(context)]


class SolidityStates(GoLikeStates):  # pylint: disable=R0903
    FUNC_KEYWORD = 'function'



'''
Language parser for Ruby-like langauges
'''

from .code_reader import CodeReader, CodeStateMachine
from .script_language import ScriptLanguageMixIn


def state_embedded_doc(token):
    if token == "=end":
        return True


class RubylikeStateMachine(CodeStateMachine):
    # pylint: disable=R0903
    FUNC_KEYWORD = 'def'

    def __init__(self, context):
        super(RubylikeStateMachine, self).__init__(context)

    def _state_global(self, token):
        if token in ("end", "}"):
            self.statemachine_return()
        elif token == self.FUNC_KEYWORD:
            self._state = self._def
            self.next(self._def)
        elif token == 'it':
            self._state = self._it
        elif token in ("begin", "do", "class", "module", "{", "${"
                       ) and self.last_token != ".":
            self.sub_state(self.statemachine_clone())
        elif token in ("while", "for"):
            if self.is_newline():
                self.next(self._for_while)
        elif token in ("if", "unless"):
            if self.is_newline():
                self.sub_state(self.statemachine_clone())
            else:
                self.next(self._if)
        elif token == "=begin":
            self.sub_state(state_embedded_doc)

    def is_newline(self):
        return self.context.newline or self.last_token == ";"

    def _def(self, token):
        if token == '(':
            self.context.push_new_function('(anonymous)')
            self.next(self._def_parameters)
            return
        self.context.push_new_function(token)
        self.next(self._def_continue)

    def _it(self, token):
        if token in ('do', '{'):
            self.context.push_new_function(self.last_token)
            self.next(self._def_continue)

    def _def_continue(self, token):
        def callback():
            self.context.end_of_function()
            self.next(self._state_global)
        if token == ".":
            self.context.add_to_function_name(token)
            self.next(self._def_class_method)
        elif token == "(":
            self.next(self._def_parameters)
        else:
            self.sub_state(self.statemachine_clone(), callback, token)

    def _def_class_method(self, token):
        self.context.add_to_function_name(token)
        self.next(self._def_continue)

    def _def_parameters(self, token):
        if token == ')':
            self.next(self._def_continue)
        else:
            self.context.parameter(token)
            return
        self.context.add_to_long_function_name(" " + token)

    def _if(self, token):
        if self.is_newline():
            self.next(self._state_global, token)
        elif token == "then":
            self.next(self._state_global)
            self.sub_state(self.statemachine_clone())

    def _for_while(self, token):
        if self.is_newline() or token == "do":
            self.next(self._state_global)
            if token != "end":
                self.sub_state(self.statemachine_clone())


class RubylikeReader(CodeReader, ScriptLanguageMixIn):
    # pylint: disable=R0903

    _conditions = set(['if', 'until', 'for', 'while', 'and', 'or',
                       'elsif', 'elseif', 'rescue',
                       'ensure', 'when', '||', '&&', '?'])

    def __init__(self, context):
        super(RubylikeReader, self).__init__(context)
        self.parallel_states = [RubylikeStateMachine(context)]



'''
Get Duplicated parameter lists
'''
from collections import Counter
from .extension_base import ExtensionBase


DEFAULT_MIN_PARAM_COUNT = 5


class xlizardExtension(ExtensionBase):
    FUNCTION_INFO = {
        "parameter_list_duplicates": {
            "caption": " dup_param_list ",
            "average_caption": " avg_dpl "},
        "parameter_list_duplicated_in_files": {
            "caption": " dup_param_list_f ",
            "average_caption": " avg_dpl_f "},
    }

    def __init__(self, context=None):
        self.all_count = Counter()
        self.all_count_per_file = Counter()
        super(xlizardExtension, self).__init__(context)

    def cross_file_process(self, fileinfos):
        saved_file_infos = []
        for fileinfo in fileinfos:
            flt = [
                    f for f in fileinfo.function_list
                    if len(f.parameters) >= DEFAULT_MIN_PARAM_COUNT
                 ]
            self.all_count.update(self._parameters(f) for f in flt)
            self.all_count_per_file.update(
                    set(self._parameters(f) for f in flt))
            for flist in fileinfo.function_list:
                flist.parameter_list_duplicates = 0
                flist.parameter_list_duplicated_in_files = 0
            yield fileinfo
            saved_file_infos.append(fileinfo)
        for fileinfo in saved_file_infos:
            self._reduce_again(fileinfo)

    def _reduce_again(self, fileinfo):
        for flist in fileinfo.function_list:
            flist.parameter_list_duplicates = self.all_count[
                self._parameters(flist)]
            flist.parameter_list_duplicated_in_files = \
                self.all_count_per_file[self._parameters(flist)]

    @staticmethod
    def _parameters(finfo):
        return ",".join(finfo.parameters)



"""
This module extends the default output formatting to include CSV.

The output is intended to be the same in structure, with additional
tokens (to reduce the need to post-process), and a reduced verbosity
due to the nature of CSV outputs. The differences are:

 * No summary of the included files, only the output
   respective of each function.
 * No additional output for functions that break any CCN
   thresholds.
 * Each line has four additional, individual tokens:
     * File name
     * Function Name
     * Function line start
     * Function line end
"""


def csv_output(result, options):
    result = result.result

    extension_variables = []
    extension_captions = []
    for extension in options.extensions:
        if extension.__class__.__name__ == 'xlizardExtension':
            if hasattr(extension, 'FUNCTION_INFO') and \
                    len(list(extension.FUNCTION_INFO)) == 1:
                extension_variable = list(extension.FUNCTION_INFO)[0]
                extension_variables.append(extension_variable)
                var_extension = extension.FUNCTION_INFO[extension_variable]
                extension_caption = var_extension['caption'] \
                    if 'caption' in var_extension else 'No_caption'
                extension_captions.append(extension_caption)

    if options.verbose:
        extension_caption = ""
        for caption in extension_captions:
            extension_caption = f"{extension_caption},{caption}"
        print(f"NLOC,CCN,token,PARAM,length,location,file,function," +
              f"long_name,start,end{extension_caption}")

    for source_file in result:
        if source_file:
            for source_function in source_file.function_list:
                if source_function:
                    extension_string = ''
                    for variable in extension_variables:
                        extension_string = f"{extension_string},{source_function.__getattribute__(variable)}"
                    print(f"{source_function.nloc},{source_function.cyclomatic_complexity},"
                          f"{source_function.token_count},{len(source_function.parameters)},"
                          f'{source_function.length},"{source_function.name.replace(chr(34), chr(39))}@'
                          f'{source_function.start_line}-{source_function.end_line}@{source_file.filename}",'
                          f'"{source_file.filename}","{source_function.name.replace(chr(34), chr(39))}",'
                          f'"{source_function.long_name.replace(chr(34), chr(39))}",'
                          f"{source_function.start_line},{source_function.end_line}{extension_string}")



'''
"JavaScript style language" includes JavaScript and PHP
'''

from .code_reader import CodeStateMachine


class JavaScriptStyleLanguageStates(CodeStateMachine):  # pylint: disable=R0903
    def __init__(self, context):
        super(JavaScriptStyleLanguageStates, self).__init__(context)
        self.last_tokens = ''
        self.function_name = ''
        self.started_function = None
        self.as_object = False
        self._getter_setter_prefix = None
        self.arrow_function_pending = False

    def _state_global(self, token):
        if self.as_object:
            # Support for getter/setter: look for 'get' or 'set' before method name
            if token in ('get', 'set'):
                self._getter_setter_prefix = token
                return
            if hasattr(self, '_getter_setter_prefix') and self._getter_setter_prefix:
                # Next token is the property name
                self.last_tokens = f"{self._getter_setter_prefix} {token}"
                self._getter_setter_prefix = None
                return
            if token == '[':
                self._collect_computed_name()
                return
            if token == ':':
                self.function_name = self.last_tokens
                return
            elif token == '(':
                if not self.started_function:
                    self.arrow_function_pending = True
                    self._function(self.last_tokens)
                self.next(self._function, token)
                return

        if token in '.':
            self._state = self._field
            self.last_tokens += token
            return
        if token == 'function':
            self._state = self._function
        elif token in ('if', 'switch', 'for', 'while', 'catch'):
            self.next(self._expecting_condition_and_statement_block)
        elif token in ('else', 'do', 'try', 'final'):
            self.next(self._expecting_statement_or_block)
        elif token in ('=>',):
            # Only handle arrow function body, do not push function here
            self._state = self._arrow_function
        elif token == '=':
            self.function_name = self.last_tokens
        elif token == "(":
            self.sub_state(
                self.__class__(self.context))
        elif token in '{':
            if self.started_function:
                self.sub_state(
                    self.__class__(self.context),
                    self._pop_function_from_stack)
            else:
                self.read_object()
        elif token in ('}', ')'):
            self.statemachine_return()
        elif self.context.newline or token == ';':
            self.function_name = ''
            self._pop_function_from_stack()

        self.last_tokens = token

    def read_object(self):
        def callback():
            self.next(self._state_global)

        object_reader = self.__class__(self.context)
        object_reader.as_object = True
        self.sub_state(object_reader, callback)

    def statemachine_before_return(self):
        self._pop_function_from_stack()

    def _expecting_condition_and_statement_block(self, token):
        def callback():
            self.next(self._expecting_statement_or_block)

        if token == "await":
            return

        if token != '(':
            self.next(self._state_global, token)
            return

        self.sub_state(
            self.__class__(self.context), callback)

    def _expecting_statement_or_block(self, token):
        def callback():
            self.next(self._state_global)
        if token == "{":
            self.sub_state(
                self.__class__(self.context), callback)
        else:
            self.next(self._state_global, token)

    def _push_function_to_stack(self):
        self.started_function = True
        self.context.push_new_function(self.function_name or '(anonymous)')

    def _pop_function_from_stack(self):
        if self.started_function:
            self.context.end_of_function()
        self.started_function = None

    def _arrow_function(self, token):
        self._push_function_to_stack()
        # Handle arrow function body
        if token == '{':
            # Block body
            self.next(self._state_global, token)
        else:
            # Expression body
            self.next(self._state_global, token)

    def _function(self, token):
        if token == '*':
            return
        if token != '(':
            self.function_name = token
        else:
            if not self.started_function:
                self._push_function_to_stack()
            self.arrow_function_pending = False
            self._state = self._dec
            if token == '(':
                self._dec(token)

    def _field(self, token):
        self.last_tokens += token
        self._state = self._state_global

    def _dec(self, token):
        if token == ')':
            self._state = self._expecting_func_opening_bracket
        elif token != '(':
            self.context.parameter(token)
            return
        self.context.add_to_long_function_name(" " + token)

    def _expecting_func_opening_bracket(self, token):
        # Do not reset started_function for arrow functions (=>)
        if token != '{' and token != '=>':
            self.started_function = None
        self.next(self._state_global, token)

    def _collect_computed_name(self):
        # Collect tokens between [ and ]
        tokens = []

        def collect(token):
            if token == ']':
                # Try to join tokens and camelCase if possible
                name = ''.join(tokens)
                # Remove quotes and pluses for simple cases
                name = name.replace("'", '').replace('"', '').replace('+', '').replace(' ', '')
                # Lowercase first char, uppercase next word's first char
                name = self._to_camel_case(name)
                self.last_tokens = name
                self.next(self._state_global)
                return True
            tokens.append(token)
            return False
        self.next(collect)

    def _to_camel_case(self, s):
        # Simple camelCase conversion for test case
        if not s:
            return s
        parts = s.split()
        if not parts:
            return s
        return parts[0][0].lower() + parts[0][1:] + ''.join(p.capitalize() for p in parts[1:])



'''
Language parser for Go lang
'''

from .code_reader import CodeStateMachine


class GoLikeStates(CodeStateMachine):  # pylint: disable=R0903

    FUNC_KEYWORD = 'func'

    def _state_global(self, token):
        if token == self.FUNC_KEYWORD:
            self._state = self._function_name
            self.context.push_new_function('')
        elif token == 'type':
            self._state = self._type_definition
        elif token in '{':
            self.sub_state(self.statemachine_clone())
        elif token in '}':
            self.statemachine_return()

    def _type_definition(self, token):
        self._state = self._after_type_name

    def _after_type_name(self, token):
        if token == 'struct':
            self._state = self._struct_definition
        elif token == 'interface':
            self._state = self._interface_definition
        else:
            self._state = self._state_global

    @CodeStateMachine.read_inside_brackets_then("{}", "_state_global")
    def _struct_definition(self, tokens):
        pass

    @CodeStateMachine.read_inside_brackets_then("{}", "_state_global")
    def _interface_definition(self, tokens):
        pass

    def _function_name(self, token):
        if token != '`':
            if token == '(':
                if len(self.context.stacked_functions) > 0\
                        and self.context.stacked_functions[-1].name != '*global*':
                    return self.next(self._function_dec, token)
                else:
                    return self.next(self._member_function, token)
            if token == '{':
                return self.next(self._expect_function_impl, token)
            self.context.add_to_function_name(token)
            self._state = self._expect_function_dec

    def _expect_function_dec(self, token):
        if token == '(':
            self.next(self._function_dec, token)
        elif token == "<":
            self.next(self._generalize, token)
        else:
            self._state = self._state_global

    @CodeStateMachine.read_inside_brackets_then("<>", "_expect_function_dec")
    def _generalize(self, tokens):
        pass

    @CodeStateMachine.read_inside_brackets_then("()", '_function_name')
    def _member_function(self, tokens):
        self.context.add_to_long_function_name(tokens)

    @CodeStateMachine.read_inside_brackets_then("()", '_expect_function_impl')
    def _function_dec(self, token):
        if token not in '()':
            self.context.parameter(token)

    def _expect_function_impl(self, token):
        if token == '{' and self.last_token != 'interface':
            self.next(self._function_impl, token)

    def _function_impl(self, _):
        def callback():
            self._state = self._state_global
            self.context.end_of_function()
        self.sub_state(self.statemachine_clone(), callback)



'''
This is an extension of xlizard. It add a list of language keywords
that adding the complexity and the line numbers of the keywords appear.
'''


class xlizardExtension(object):  # pylint: disable=R0903

    # pylint: disable=W0221
    def __call__(self, tokens, reader):
        context = reader.context
        conditions = reader.conditions
        for token in tokens:
            yield token
            if not hasattr(context.current_function, 'complex_tags'):
                context.current_function.complex_tags = []
            if token in conditions:
                context.current_function.complex_tags.append(
                        [token, context.current_line])



"""
This is an extension of xlizard, that counts the statements in a function
"""


class xlizardExtension:  # pylint: disable=R0903

    FUNCTION_INFO = {"statement_count": {"caption": "statements"}}

    def __call__(self, tokens, reader):
        c_family = 'c' in reader.language_names or \
                   'cpp' in reader.language_names

        block_count = 0
        for token in tokens:
            if c_family:
                if not hasattr(reader.context.current_function,
                               "statement_count"):
                    reader.context.current_function.statement_count = 0
                if token in [';', 'if', 'for', 'while', ':', 'switch']:
                    reader.context.current_function.statement_count += 1
                if token == "{":
                    if block_count != 0:
                        reader.context.\
                            current_function.statement_count += 1
                    block_count += 1
                if token == "}":
                    block_count -= 1 if block_count > 0 else 0
            else:
                reader.context.current_function.statement_count = None

            yield token



'''
This is an extension of xlizard. It use a non strict standard to count
the cyclomatic complexity number, where the boolean operators (
&&, ||, and, or, not, !) would be counted.
'''


class xlizardExtension(object):  # pylint: disable=R0903

    # pylint: disable=W0221
    def __call__(self, tokens, reader):
        reader.conditions -= set(['&&', '||', 'and', 'or'])
        return tokens



'''
Language parser for Perl
'''

from .code_reader import CodeReader, CodeStateMachine
from .script_language import ScriptLanguageMixIn


class PerlCommentsMixin(object):
    @staticmethod
    def get_comment_from_token(token):
        if token.startswith('#'):
            return token  # Return the entire comment including #
        return None


class PerlReader(CodeReader, ScriptLanguageMixIn):
    # pylint: disable=R0903

    ext = ['pl', 'pm']
    language_names = ['perl']
    _conditions = set(['if', 'elsif', 'unless', 'while', 'until', 'for', 'foreach',
                       '&&', '||', '?', ':', 'when', 'given', 'default', 'do'])

    def __init__(self, context):
        super(PerlReader, self).__init__(context)
        self.parallel_states = [PerlStates(context)]

    def preprocess(self, tokens):
        comment = None
        for token in tokens:
            if comment is not None:
                if token == '\n':
                    yield comment
                    comment = None
                    yield token
                else:
                    comment += token
            elif token == '#':
                comment = token
            else:
                yield token
        if comment is not None:
            yield comment

    def _state(self, token):
        current_state = self.parallel_states[-1]
        if token == '\n':
            return
        return current_state.state(token)

    @staticmethod
    def get_comment_from_token(token):
        if token.startswith('#'):
            # For forgiveness comments, return the entire comment
            stripped = token.lstrip('#').strip()
            if stripped.startswith('xlizard forgives') or stripped.startswith('#xlizard forgives'):
                return '#xlizard forgives'  # Return standardized forgiveness comment
            return stripped  # Return the stripped comment for other case
        return None

    @staticmethod
    def generate_tokens(source_code, addition='', token_class=None):
        return ScriptLanguageMixIn.generate_common_tokens(source_code, addition, token_class)


class PerlStates(CodeStateMachine):
    _conditions = set(['if', 'elsif', 'unless', 'while', 'until', 'for', 'foreach',
                       '&&', '||', '?', ':', 'when', 'given', 'default', 'do'])

    def __init__(self, context):
        super(PerlStates, self).__init__(context)
        self.function_name = ''
        self.package_name = ''
        self.variable_name = ''
        self.brace_count = 0
        self.paren_count = 0
        self.in_attribute = False
        self.anonymous_count = 0
        self._state = self._state_global

    def _state_global(self, token):
        if token == 'package':
            self.next(self._state_package_dec)
        elif token == 'sub':
            self.function_name = ''
            self.next(self._state_function_dec)
        elif token == '{':
            self.brace_count += 1
        elif token == '}':
            self.brace_count -= 1
        elif token == '(':
            self.paren_count += 1
            self.next(self._state_function_call)
        elif token in ('$', 'my', 'our', 'local'):
            self.variable_name = ''
            self.next(self._state_variable)

    def _state_package_dec(self, token):
        if not token.isspace():
            self.package_name = token
            self.next(self._state_global)

    def _state_variable(self, token):
        if token == '$':
            # Skip the $ in variable name
            pass
        elif token == '=':
            self.next(self._state_assignment)
        elif token == ';':
            self.variable_name = ''
            self.next(self._state_global)
        elif not token.isspace() and self.variable_name == '':
            self.variable_name = token

    def _state_assignment(self, token):
        if token == 'sub':
            self.next(self._state_anon_sub)
        elif token == ';':
            self.variable_name = ''
            self.next(self._state_global)

    def _state_function_call(self, token):
        if token == 'sub':
            # Inline anonymous subroutine as argument
            self.anonymous_count += 1
            full_name = "<anonymous>"
            if self.package_name:
                full_name = f"{self.package_name}::{full_name}"

            self.context.try_new_function(full_name)
            self.context.confirm_new_function()
            self.next(self._state_anon_brace_search)
        elif token == ')':
            self.paren_count -= 1
            if self.paren_count == 0:
                self.next(self._state_global)
        elif token == '(':
            self.paren_count += 1

    def _state_anon_sub(self, token):
        if token == '{':
            self.brace_count = 1
            full_name = '<anonymous>'
            # Use variable name if available for more readable function name
            if self.variable_name:
                full_name = '$' + self.variable_name

            if self.package_name:
                full_name = f"{self.package_name}::{full_name}"

            self.context.try_new_function(full_name)
            self.context.confirm_new_function()
            self.next(self._state_function_body)

    def _state_function_dec(self, token):
        if token == '{':
            self.brace_count = 1
            if self.function_name:
                full_name = self.function_name
                if self.package_name:
                    full_name = f"{self.package_name}::{self.function_name}"
                self.context.try_new_function(full_name)
                self.context.confirm_new_function()
            self.next(self._state_function_body)
        elif token == ':':
            self.in_attribute = True
        elif token == ';':
            # Forward declaration like "sub func_name;"
            if self.function_name:
                full_name = self.function_name
                if self.package_name:
                    full_name = f"{self.package_name}::{self.function_name}"
                self.context.try_new_function(full_name)
                self.context.confirm_new_function()
                # Empty function body
                self.context.end_of_function()
            self.next(self._state_global)
        elif token == '(':
            # Function with parameter prototype like "sub fetch($)"
            # Save the current function name and look for closing paren
            self.paren_count = 1
            self.next(self._state_function_prototype)
        elif token == 'sub':
            # Handle anonymous subroutine like 'callback(sub { ... })'
            self.anonymous_count += 1
            full_name = "<anonymous>"
            if self.package_name:
                full_name = f"{self.package_name}::{full_name}"
            self.context.try_new_function(full_name)
            self.context.confirm_new_function()
            self.next(self._state_anon_brace_search)
        elif not token.isspace():
            if not self.in_attribute:
                self.function_name = token
            else:
                # Skip attribute name
                self.in_attribute = False

    def _state_function_prototype(self, token):
        # Handle parameter prototypes after function name: sub fetch($) { ... }
        if token == ')':
            self.paren_count -= 1
            if self.paren_count == 0:
                # Return to function declaration state to handle the opening brace
                self.next(self._state_function_dec)
        elif token == '(':
            self.paren_count += 1

    def _state_anon_brace_search(self, token):
        if token == '{':
            self.brace_count = 1
            self.next(self._state_function_body)
        elif token == '(':
            self.paren_count += 1
        elif token == ')':
            self.paren_count -= 1
            if self.paren_count == 0:
                self.next(self._state_global)

    def _state_function_body(self, token):
        if token == '{':
            self.brace_count += 1
        elif token == '}':
            self.brace_count -= 1
            if self.brace_count == 0:
                self.context.end_of_function()
                self.next(self._state_global)
        elif token == '?':
            # Ternary operator increases complexity
            self.context.add_condition()
        elif token == ':':
            # Colon part of ternary operator also increases complexity
            self.context.add_condition()
        elif token == 'sub':
            # Check if it's a nested named subroutine or anonymou
            self.next(self._state_nested_sub_dec)
        elif token == '(':
            # Track function calls inside function body
            self.paren_count += 1
            self.next(self._state_nested_call)

    def _state_nested_sub_dec(self, token):
        if token.isspace():
            return
        if token == '{':
            # Anonymous sub
            self.brace_count += 1
            self.anonymous_count += 1
            anon_name = "<anonymous>"
            if self.package_name:
                anon_name = f"{self.package_name}::{anon_name}"
            self.context.add_condition()  # Count sub as a condition
            self.next(self._state_function_body)
        else:
            # Named nested sub
            nested_func_name = token
            full_name = nested_func_name
            if self.package_name:
                full_name = f"{self.package_name}::{nested_func_name}"

            # Create a new function for the nested sub
            self.context.try_new_function(full_name)
            self.context.confirm_new_function()
            self.next(self._state_nested_named_sub_brace_search)

    def _state_nested_named_sub_brace_search(self, token):
        if token == '{':
            self.brace_count = 1
            self.next(self._state_nested_sub_body)
        elif token.isspace():
            return
        elif token == ':':
            # Handle attributes in nested sub
            self.in_attribute = True
            return
        elif token == ';':
            # Forward declaration
            self.context.end_of_function()
            self.next(self._state_function_body)

    def _state_nested_sub_body(self, token):
        if token == '{':
            self.brace_count += 1
        elif token == '}':
            self.brace_count -= 1
            if self.brace_count == 0:
                # End the nested function
                self.context.end_of_function()
                # Return to parent function
                self.next(self._state_function_body)

    def _state_nested_call(self, token):
        if token == 'sub':
            # Inline anonymous subroutine as argument
            self.anonymous_count += 1
            full_name = "<anonymous>"
            if self.package_name:
                full_name = f"{self.package_name}::{full_name}"

            self.context.try_new_function(full_name)
            self.context.confirm_new_function()
            self.next(self._state_nested_anon_search)
        elif token == ')':
            self.paren_count -= 1
            if self.paren_count == 0:
                self.next(self._state_function_body)
        elif token == '(':
            self.paren_count += 1

    def _state_nested_anon_search(self, token):
        if token == '{':
            self.brace_count += 1
            self.next(self._state_nested_anon_body)
        elif token == '(':
            self.paren_count += 1
        elif token == ')':
            self.paren_count -= 1
            if self.paren_count == 0:
                self.next(self._state_function_body)

    def _state_nested_anon_body(self, token):
        if token == '{':
            self.brace_count += 1
        elif token == '}':
            self.brace_count -= 1
            if self.brace_count == 1:  # Back to outer function level
                self.context.end_of_function()
                self.next(self._state_function_body)



''' Language parser for Python '''

from .code_reader import CodeReader, CodeStateMachine
from .script_language import ScriptLanguageMixIn


def count_spaces(token):
    return len(token.replace('\t', ' ' * 8))


class PythonIndents:  # pylint: disable=R0902
    def __init__(self, context):
        self.indents = [0]
        self.context = context

    def set_nesting(self, spaces, token=""):
        while self.indents[-1] > spaces and (not token.startswith(")")):
            self.indents.pop()
            self.context.pop_nesting()
        if self.indents[-1] < spaces:
            self.indents.append(spaces)
            self.context.add_bare_nesting()

    def reset(self):
        self.set_nesting(0)


class PythonReader(CodeReader, ScriptLanguageMixIn):

    ext = ['py']
    language_names = ['python']
    _conditions = set([
        'if', 'for', 'while', 'and', 'or',
        'elif', 'except', 'finally'
    ])

    def __init__(self, context):
        super(PythonReader, self).__init__(context)
        self.parallel_states = [PythonStates(context, self)]

    @staticmethod
    def generate_tokens(source_code, addition='', token_class=None):
        return ScriptLanguageMixIn.generate_common_tokens(
            source_code,
            r"|(?:\"\"\"(?:\\.|[^\"]|\"(?!\"\")|\"\"(?!\"))*\"\"\")" +
            r"|(?:\'\'\'(?:\\.|[^\']|\'(?!\'\')|\'\'(?!\'))*\'\'\')",
            token_class)

    def preprocess(self, tokens):
        indents = PythonIndents(self.context)
        current_leading_spaces = 0
        reading_leading_space = True
        for token in tokens:
            if token != '\n':
                if reading_leading_space:
                    if token.isspace():
                        current_leading_spaces += count_spaces(token)
                    else:
                        if not token.startswith('#'):
                            current_function = self.context.current_function
                            if (current_function.name == '*global*' or
                                    current_function.long_name.endswith(')')):
                                indents.set_nesting(current_leading_spaces, token)
                        reading_leading_space = False
            else:
                reading_leading_space = True
                current_leading_spaces = 0
            if not token.isspace() or token == '\n':
                yield token
        indents.reset()


class PythonStates(CodeStateMachine):  # pylint: disable=R0903
    def __init__(self, context, reader):
        super(PythonStates, self).__init__(context)
        self.reader = reader

    def _state_global(self, token):
        if token == 'def':
            self._state = self._function

    def _function(self, token):
        if token != '(':
            self.context.restart_new_function(token)
            self.context.add_to_long_function_name("(")
        else:
            self._state = self._dec

    def _dec(self, token):
        if token == ')':
            self._state = self._state_colon
        elif token == '[':
            self._state = self._state_parameterized_type_annotation
        else:
            self.context.parameter(token)
            return
        self.context.add_to_long_function_name(" " + token)

    def _state_colon(self, token):
        if token == ':':
            self.next(self._state_first_line)
        else:
            self.next(self._state_global)

    def _state_first_line(self, token):
        self._state = self._state_global
        if token.startswith('"""') or token.startswith("'''"):
            self.context.add_nloc(-token.count('\n') - 1)
        self._state_global(token)

    def _state_parameterized_type_annotation(self, token):
        self.context.add_to_long_function_name(" " + token)
        if token == ']':
            self._state = self._dec



'''
Language parser for JSX
'''

from .javascript import JavaScriptReader
from .typescript import JSTokenizer, Tokenizer, TypeScriptStates
from .code_reader import CodeReader
from .js_style_regex_expression import js_style_regex_expression
from .js_style_language_states import JavaScriptStyleLanguageStates


class JSXTypeScriptStates(TypeScriptStates):
    """State machine for JSX/TSX files extending TypeScriptStates"""

    def __init__(self, context):
        super().__init__(context)
        # Initialize attributes that might be accessed later
        self._parent_function_name = None
        self.in_variable_declaration = False
        self.last_variable_name = None

    def statemachine_before_return(self):
        # Ensure the main function is closed at the end
        if self.started_function:
            self._pop_function_from_stack()
            # After popping, if current_function is not *global*, pop again to add to function_list
            if self.context.current_function and self.context.current_function.name != "*global*":
                self.context.end_of_function()

    def _state_global(self, token):
        # Handle variable declarations
        if token in ('const', 'let', 'var'):
            self.in_variable_declaration = True
            super()._state_global(token)
            return

        if self.in_variable_declaration:
            if token == '=':
                # Save the variable name when we see the assignment
                self.last_variable_name = self.last_tokens.strip()
                super()._state_global(token)
                return
            elif token == '=>':
                # We're in an arrow function with a variable assignment
                if self.last_variable_name and not self.started_function:
                    self.function_name = self.last_variable_name
                    self._push_function_to_stack()
                    self.in_variable_declaration = False
                    # Switch to arrow function state to handle the body
                    self._state = self._arrow_function
                    return
            elif token == ';' or self.context.newline:
                self.in_variable_declaration = False

        # Handle arrow function in JSX/TSX prop context
        if token == '=>' and not self.in_variable_declaration:
            if not self.started_function:
                self.function_name = '(anonymous)'
                self._push_function_to_stack()
                return

        if not self.as_object:
            if token == ':':
                self._consume_type_annotation()
                return

        # Pop anonymous function after closing '}' in TSX/JSX prop
        if token == '}' and self.started_function and self.function_name == '(anonymous)':
            self._pop_function_from_stack()

        # Continue with regular TypeScript state handling
        super()._state_global(token)

    def _arrow_function(self, token):
        self.next(self._state_global, token)


class TSXTokenizer(JSTokenizer):
    def __init__(self):
        super().__init__()

    def process_token(self, token):
        if token == "<":
            from .jsx import XMLTagWithAttrTokenizer  # Import only when needed
            self.sub_tokenizer = XMLTagWithAttrTokenizer()
            return

        if token == "=>":
            # Special handling for arrow functions
            yield token
            return

        for tok in super().process_token(token):
            yield tok


class JSXMixin:
    '''Base mixin class for JSX/TSX shared functionality'''
    @staticmethod
    @js_style_regex_expression
    def generate_tokens(source_code, addition='', token_class=None):
        addition = addition +\
            r"|(?:\$\w+)" + \
            r"|(?:\<\/\w+\>)" + \
            r"|(?:=>)" + \
            r"|`.*?`"
        js_tokenizer = TSXTokenizer()
        for token in CodeReader.generate_tokens(
                source_code, addition, token_class):
            for tok in js_tokenizer(token):
                yield tok

    def _expecting_func_opening_bracket(self, token):
        if token == '<':
            self.next(self._expecting_jsx)
            return
        if token == '=>':
            # Handle arrow functions in JSX attributes
            self._handle_arrow_function()
            return
        super()._expecting_func_opening_bracket(token)

    def _handle_arrow_function(self):
        # Process arrow function in JSX context
        self.context.add_to_long_function_name(" => ")

        # Store the current function
        current_function = self.context.current_function

        # Create a new anonymous function
        self.context.restart_new_function('(anonymous)')

        # Set up for the arrow function body
        def callback():
            # Return to the original function when done
            self.context.current_function = current_function

        self.sub_state(self.__class__(self.context), callback)

    def _expecting_arrow_function_body(self, token):
        if token == '{':
            # Arrow function with block body
            self.next(self._function_body)
        else:
            # Arrow function with expression body
            self.next(self._expecting_func_opening_bracket)

    def _function_body(self, token):
        if token == '}':
            # End of arrow function body
            self.context.end_of_function()
            self.next(self._expecting_func_opening_bracket)

    def _expecting_jsx(self, token):
        if token == '>':
            self.next(self._expecting_func_opening_bracket)


class JSXJavaScriptStyleLanguageStates(JavaScriptStyleLanguageStates):
    def __init__(self, context):
        super(JSXJavaScriptStyleLanguageStates, self).__init__(context)

    def _state_global(self, token):
        # Handle variable declarations
        if token in ('const', 'let', 'var'):
            # Remember that we're in a variable declaration
            self.in_variable_declaration = True
            super()._state_global(token)
            return

        if hasattr(self, 'in_variable_declaration') and self.in_variable_declaration:
            if token == '=':
                # We're in a variable assignment
                self.function_name = self.last_tokens.strip()
                super()._state_global(token)
                return
            elif token == '=>':
                # We're in an arrow function with a variable assignment
                if not self.started_function and self.function_name:
                    self._push_function_to_stack()
                self._state = self._arrow_function
                return
            elif token == ';' or self.context.newline:
                # End of variable declaration
                self.in_variable_declaration = False

        super()._state_global(token)

    def _expecting_func_opening_bracket(self, token):
        if token == ':':
            # Handle type annotations like TypeScript does
            self._consume_type_annotation()
            return
        super()._expecting_func_opening_bracket(token)

    def _consume_type_annotation(self):
        # Skip over type annotations (simplified version of TypeScript's behavior)
        def skip_until_terminator(token):
            if token in ['{', '=', ';', ')', '(', '=>']:
                self.next(self._state_global, token)
                return True
            return False

        self.next(skip_until_terminator)


class JSXReader(JavaScriptReader, JSXMixin):
    # pylint: disable=R0903

    ext = ['jsx']
    language_names = ['jsx']

    @staticmethod
    @js_style_regex_expression
    def generate_tokens(source_code, addition='', token_class=None):
        # Add support for JSX syntax patterns
        addition = addition + \
            r"|(?:<[A-Za-z][A-Za-z0-9]*(?:\.[A-Za-z][A-Za-z0-9]*)*>)" + \
            r"|(?:<\/[A-Za-z][A-Za-z0-9]*(?:\.[A-Za-z][A-Za-z0-9]*)*>)"
        return JSXMixin.generate_tokens(source_code, addition, token_class)

    def __init__(self, context):
        super(JSXReader, self).__init__(context)
        # Use our JSXTypeScriptStates for better handling of JSX
        self.parallel_states = [JSXTypeScriptStates(context)]


class XMLTagWithAttrTokenizer(Tokenizer):
    def __init__(self):
        super(XMLTagWithAttrTokenizer, self).__init__()
        self.tag = None
        self.state = self._global_state
        self.cache = ['<']
        self.brace_count = 0  # Track nested braces for complex expressions
        self.arrow_function_detected = False  # Track if we've detected an arrow function

    def process_token(self, token):
        self.cache.append(token)
        if not token.isspace():
            result = self.state(token)
            if result is not None:
                if isinstance(result, list):
                    for tok in result:
                        yield tok
                else:
                    return result
        return ()

    def abort(self):
        self.stop()
        return self.cache

    def flush(self):
        tmp, self.cache = self.cache, []
        return [''.join(tmp)]

    def _global_state(self, token):
        if not isidentifier(token):
            return self.abort()
        self.tag = token
        self.state = self._after_tag

    def _after_tag(self, token):
        if token == '>':
            self.state = self._body
        elif token == "/":
            self.state = self._expecting_self_closing
        elif isidentifier(token):
            self.state = self._expecting_equal_sign
        else:
            return self.abort()

    def _expecting_self_closing(self, token):
        if token == ">":
            self.stop()
            return self.flush()
        return self.abort()

    def _expecting_equal_sign(self, token):
        if token == '=':
            self.state = self._expecting_value
        else:
            return self.abort()

    def _expecting_value(self, token):
        if token[0] in "'\"":
            self.state = self._after_tag
        elif token == "{":
            self.brace_count = 1  # Start counting braces
            self.state = self._jsx_expression
            # Don't add the closing brace automatically
            # self.cache.append("}")
            self.sub_tokenizer = TSXTokenizer()

    def _jsx_expression(self, token):
        # Handle nested braces in expressions
        if token == "{":
            self.brace_count += 1
        elif token == "}":
            self.brace_count -= 1
            if self.brace_count == 0:
                # We've found the matching closing brace
                self.state = self._after_tag
                return

        # Handle arrow functions in JSX attributes
        if token == "=>":
            self.arrow_function_detected = True
            # Explicitly yield the arrow token to ensure it's processed
            return ["=>"]

        # Handle type annotations in JSX attributes
        if token == "<":
            # This might be a TypeScript generic type annotation
            # We'll continue in the current state and let the tokenizer handle it
            pass

    def _body(self, token):
        if token == "<":
            self.sub_tokenizer = XMLTagWithAttrTokenizer()
            self.cache.pop()
            return self.flush()

        if token.startswith("</"):
            self.stop()
            return self.flush()

        if token == '{':
            self.sub_tokenizer = TSXTokenizer()
            return self.flush()


def isidentifier(token):
    try:
        return token.isidentifier()
    except AttributeError:
        return token.encode(encoding='UTF-8')[0].isalpha()



'''
Common behaviours of script languages
'''
from .code_reader import CodeReader


class ScriptLanguageMixIn:
    # pylint: disable=R0903

    @staticmethod
    def get_comment_from_token(token):
        if token.startswith("#"):
            # For forgiveness comments, return the entire comment with directive intact
            stripped = token.lstrip('#').strip()

            # Handle forgiveness directives with proper formatting
            if stripped.startswith('xlizard forgive global') or stripped.startswith('#xlizard forgive global'):
                return '#xlizard forgive global'  # Preserve global directive
            elif stripped.startswith('xlizard forgive') or stripped.startswith('#xlizard forgive'):
                return '#xlizard forgive'  # Return standardized forgiveness comment

            return stripped  # Return the stripped comment for other cases
        return None

    @staticmethod
    def generate_common_tokens(source_code, addition, match_holder=None):
        _until_end = r"(?:\\\n|[^\n])*"
        return CodeReader.generate_tokens(
            source_code,
            r"|\#" + _until_end + addition,
            match_holder)



__version__ = "1.0.10"  # Должно совпадать с версией в setup.py
version = "1.0.10"      # Для обратной совместимости

# Явно экспортируем нужные модули
from .xlizard import *
from xlizard.combined_metrics import CombinedMetrics
from xlizard.sourcemonitor_metrics import SourceMonitorMetrics

__all__ = ['CombinedMetrics', 'load_thresholds', 'DEFAULT_THRESHOLDS', 'Config', 'SourceMonitorMetrics']


'''
Language parser for JavaScript
'''

import re
from .code_reader import CodeReader, CodeStateMachine
from .clike import CCppCommentsMixin
from .js_style_language_states import JavaScriptStyleLanguageStates
from .js_style_regex_expression import js_style_regex_expression


class Tokenizer(object):
    def __init__(self):
        self.sub_tokenizer = None
        self._ended = False

    def __call__(self, token):
        if self.sub_tokenizer:
            for tok in self.sub_tokenizer(token):
                yield tok
            if self.sub_tokenizer._ended:
                self.sub_tokenizer = None
            return
        for tok in self.process_token(token):
            yield tok

    def stop(self):
        self._ended = True

    def process_token(self, token):
        pass


class JSTokenizer(Tokenizer):
    def __init__(self):
        super().__init__()
        self.depth = 1

    def process_token(self, token):
        if token == "{":
            self.depth += 1
        elif token == "}":
            self.depth -= 1
            if self.depth == 0:
                self.stop()
                return
        yield token


class TypeScriptReader(CodeReader, CCppCommentsMixin):
    # pylint: disable=R0903

    ext = ['ts']
    language_names = ['typescript', 'ts']
    _conditions = set(['if', 'elseif', 'for', 'while', '&&', '||', '?',
                       'catch', 'case'])

    def __init__(self, context):
        super().__init__(context)
        self.parallel_states = [TypeScriptStates(context)]

    @staticmethod
    @js_style_regex_expression
    def generate_tokens(source_code, addition='', token_class=None):
        def split_template_literal(token, quote):
            content = token[1:-1]
            i = 0
            # Special case for double-quoted strings starting with ${
            if quote == '"' and content.startswith('${'):
                yield '""'
            while i < len(content):
                idx = content.find('${', i)
                if idx == -1:
                    if i < len(content):
                        yield quote + content[i:] + quote
                    break
                if idx > i and not (quote == '"' and idx == 0 and content.startswith('${')):
                    yield quote + content[i:idx] + quote
                yield '${'
                i = idx + 2
                expr_start = i
                brace_count = 1
                while i < len(content) and brace_count > 0:
                    if content[i] == '{':
                        brace_count += 1
                    elif content[i] == '}':
                        brace_count -= 1
                    i += 1
                expr = content[expr_start:i-1]
                yield expr
                yield '}'
                content = content[i:]
                i = 0
        # Restore original addition pattern for template literals
        addition = addition + r"|(?:\$\w+)" + r"|(?:\w+\?)" + r"|`.*?`"
        for token in CodeReader.generate_tokens(source_code, addition, token_class):
            if (
                isinstance(token, str)
                and (token.startswith('`') or token.startswith('"'))
                and token[0] == token[-1]
                and '${' in token
            ):
                quote = token[0]
                for t in split_template_literal(token, quote):
                    yield t
                continue
            yield token


class TypeScriptStates(JavaScriptStyleLanguageStates):
    def __init__(self, context):
        super().__init__(context)

    def statemachine_before_return(self):
        # Ensure the main function is closed at the end
        if self.started_function:
            self._pop_function_from_stack()

    def _state_global(self, token):
        if not self.as_object:
            if token == ':':
                self._consume_type_annotation()
                return
        super()._state_global(token)

    def _expecting_func_opening_bracket(self, token):
        if token == ':':
            self._consume_type_annotation()
        else:
            super()._expecting_func_opening_bracket(token)

    def _consume_type_annotation(self):
        typeStates = TypeScriptTypeAnnotationStates(self.context)

        def callback():
            if typeStates.saved_token:
                self(typeStates.saved_token)
        self.sub_state(typeStates, callback)


class TypeScriptTypeAnnotationStates(CodeStateMachine):
    def __init__(self, context):
        super().__init__(context)
        self.saved_token = None

    def _state_global(self, token):
        if token == '{':
            self.next(self._inline_type_annotation, token)
        else:
            self.next(self._state_simple_type, token)

    def _state_simple_type(self, token):
        if token == '<':
            self.next(self._state_generic_type, token)
        elif token in '{=;)':
            self.saved_token = token
            self.statemachine_return()
        elif token == '(':
            self.next(self._function_type_annotation, token)
        elif token == '=>':
            # Handle arrow function after type annotation
            self.saved_token = token
            self.statemachine_return()

    @CodeStateMachine.read_inside_brackets_then("{}")
    def _inline_type_annotation(self, _):
        self.statemachine_return()

    @CodeStateMachine.read_inside_brackets_then("<>")
    def _state_generic_type(self, token):
        self.statemachine_return()

    @CodeStateMachine.read_inside_brackets_then("()")
    def _function_type_annotation(self, _):
        self.statemachine_return()



import os
import sys
import logging
import re
from concurrent.futures import ThreadPoolExecutor
import xml.etree.ElementTree as ET
from tqdm import tqdm
from collections import defaultdict
from typing import Dict, List, Optional, DefaultDict

# Настройка логгирования
logging.basicConfig(level=logging.INFO, format='%(message)s')
logger = logging.getLogger()

class Config:
    """Конфигурация анализатора"""
    EXCLUDE_DIRS = {'.git', 'venv', '__pycache__', 'include'}
    SUPPORTED_EXTENSIONS = {'.c', '.h'}
    METRICS = [
        'comment_percentage',
        'max_block_depth',
        'pointer_operations',
        'preprocessor_directives'
    ]
    THRESHOLDS = {
        'comment_percentage': 20,
        'max_block_depth': 5,
        'pointer_operations': 20,
        'preprocessor_directives': 15
    }

class FileAnalyzer:
    """Анализатор отдельных файлов"""
    @staticmethod
    def _remove_string_literals(content: str) -> str:
        """Удаляет строковые литералы для упрощения анализа"""
        return re.sub(r'"[^"]*"', '', content)

    @staticmethod
    def _count_comments(content: str) -> int:
        """Подсчёт количества строк с комментариями"""
        lines = content.split('\n')
        comment_lines = 0
        in_block_comment = False
        
        for line in lines:
            line = line.strip()
            if in_block_comment:
                comment_lines += 1
                if '*/' in line:
                    in_block_comment = False
                continue
            if line.startswith('/*'):
                comment_lines += 1
                in_block_comment = True
                if '*/' in line:
                    in_block_comment = False
            elif line.startswith('//'):
                comment_lines += 1
                
        return comment_lines

    @staticmethod
    def _calculate_block_depth(content: str, is_function: bool = False) -> int:
        """Вычисление максимальной глубины вложенности для файла или функции"""
        max_depth = 0
        current_depth = 0
        in_function = not is_function  # Для анализа функции начинаем сразу с глубины 0
        
        for char in content:
            if char == '{':
                current_depth += 1
                if in_function:
                    max_depth = max(max_depth, current_depth)
            elif char == '}':
                if in_function:
                    current_depth -= 1
                if current_depth == 0:
                    in_function = not is_function  # Выходим из функции
                    
        return max_depth

    @classmethod
    def analyze(cls, file_path: str) -> Optional[Dict[str, float]]:
        """Анализ файла и возврат метрик"""
        try:
            with open(file_path, 'r', encoding='utf-8', errors='ignore') as f:
                content = f.read()
            
            content_no_strings = cls._remove_string_literals(content)
            total_lines = len(content.split('\n'))
            comment_lines = cls._count_comments(content)
            
            return {
                'file_name': os.path.basename(file_path),
                'file_path': os.path.relpath(file_path),
                'comment_percentage': (comment_lines / total_lines * 100) if total_lines else 0,
                'max_block_depth': cls._calculate_block_depth(content_no_strings),
                'pointer_operations': content_no_strings.count('*') + content_no_strings.count('&'),
                'preprocessor_directives': len([l for l in content.split('\n') if l.strip().startswith('#')]),
            }
        except Exception as e:
            logger.warning(f"Ошибка анализа {file_path}: {str(e)}")
            return None

class MetricsAggregator:
    """Агрегация метрик по всем файлам"""
    def __init__(self):
        self.file_metrics: List[Dict[str, float]] = []
        self.total_metrics: DefaultDict[str, float] = defaultdict(float)
        self.counts: DefaultDict[str, int] = defaultdict(int)

    def add_file_metrics(self, metrics: Dict[str, float]) -> None:
        """Добавление метрик файла"""
        self.file_metrics.append(metrics)
        for key in Config.METRICS:
            if key in metrics:
                self.total_metrics[key] += metrics[key]
                self.counts[key] += 1

    def get_averages(self) -> Dict[str, float]:
        """Расчёт средних значений"""
        return {
            metric: self.total_metrics[metric] / self.counts[metric]
            for metric in Config.METRICS
            if self.counts[metric] > 0
        }

class ReportGenerator:
    """Генерация отчётов"""
    @staticmethod
    def generate_xml(metrics: List[Dict[str, float]], output_path: str) -> None:
        """Генерация XML отчёта"""
        # Создаем папку output если ее нет
        output_dir = os.path.dirname(output_path)
        try:
            os.makedirs(output_dir, exist_ok=True)
            logger.info(f"Папка {output_dir} успешно создана или уже существует")
        except OSError as e:
            logger.error(f"Не удалось создать папку {output_dir}: {str(e)}")
            raise

        root = ET.Element('sourcemonitor_metrics')
        for file_metrics in metrics:
            file_node = ET.SubElement(root, 'file', 
                                   name=file_metrics['file_name'],
                                   path=file_metrics['file_path'])
            for metric in Config.METRICS:
                ET.SubElement(file_node, metric).text = str(file_metrics.get(metric, 0))
        
        tree = ET.ElementTree(root)
        try:
            tree.write(output_path, encoding='utf-8', xml_declaration=True)
            logger.info(f"XML-отчёт успешно сохранён в {output_path}")
        except Exception as e:
            logger.error(f"Ошибка при сохранении отчёта в {output_path}: {str(e)}")
            raise

class SourceMonitorMetrics:
    """Основной класс анализатора"""
    def __init__(self, path: str):
        self.path = os.path.abspath(path)
        # Правильное определение корня пакета xlizard
        package_root = self._find_package_root()
        self.output_xml = os.path.join(package_root, 'output', 'sourcemonitor_metrics.xml')
        self.aggregator = MetricsAggregator()
        logger.info(f"Отчёт будет сохранён в: {self.output_xml}")

    def _find_package_root(self) -> str:
        """Находит корень пакета xlizard"""
        # Ищем папку с именем 'xlizard' в пути текущего файла
        current_dir = os.path.dirname(os.path.abspath(__file__))
        while current_dir != os.path.dirname(current_dir):  # Пока не дойдем до корня файловой системы
            if os.path.basename(current_dir) == 'xlizard':
                return current_dir
            current_dir = os.path.dirname(current_dir)
        # Если не нашли, используем текущую директорию
        logger.warning("Не удалось найти корень пакета xlizard, используется текущая директория")
        return os.getcwd()

    def _collect_files(self) -> List[str]:
        """Сбор файлов для анализа"""
        c_files = []
        for root, dirs, files in os.walk(self.path):
            dirs[:] = [d for d in dirs if d not in Config.EXCLUDE_DIRS]
            c_files.extend(
                os.path.join(root, f) 
                for f in files 
                if os.path.splitext(f)[1] in Config.SUPPORTED_EXTENSIONS
            )
        return c_files

    def get_metrics(self) -> List[Dict[str, float]]:
        """Возвращает собранные метрики для интеграции"""
        return self.aggregator.file_metrics

    def analyze_directory(self) -> None:
        """Анализ директории"""
        if not os.path.exists(self.path):
            logger.error(f"Ошибка: путь '{self.path}' не существует!")
            sys.exit(1)

        files = self._collect_files()
        logger.info(f"Найдено {len(files)} файлов для анализа...")
        
        with ThreadPoolExecutor() as executor:
            results = list(tqdm(
                executor.map(FileAnalyzer.analyze, files),
                total=len(files),
                desc="Анализ файлов"
            ))
        
        for metrics in filter(None, results):
            self.aggregator.add_file_metrics(metrics)

        try:
            ReportGenerator.generate_xml(
                self.aggregator.file_metrics, 
                self.output_xml
            )
        except Exception as e:
            logger.error(f"Не удалось сохранить отчёт: {str(e)}")
            sys.exit(1)

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Использование: python sourcemonitor_metrics.py <путь_к_директории>")
        sys.exit(1)

    analyzer = SourceMonitorMetrics(sys.argv[1])
    analyzer.analyze_directory()


'''
Checkstyle XML output for xlizard
'''

def checkstyle_output(all_result, verbose):
    result = all_result.result
    import xml.etree.ElementTree as ET

    checkstyle = ET.Element('checkstyle', version="4.3")

    for source_file in result:
        if source_file:
            file_elem = ET.SubElement(checkstyle, 'file', name=source_file.filename)
            for func in source_file.function_list:
                # Each function with a warning (e.g., CCN > threshold) could be an <error>
                # For now, output all functions as <error> for demonstration
                ET.SubElement(
                    file_elem,
                    'error',
                    line=str(func.start_line),
                    column="0",
                    severity="info",
                    message=f"{func.name} has {func.nloc} NLOC, {func.cyclomatic_complexity} CCN, {func.token_count} token, {len(func.parameters)} PARAM, {func.length} length",
                    source="xlizard"
                )

    # Pretty print
    import xml.dom.minidom
    rough_string = ET.tostring(checkstyle, 'utf-8')
    reparsed = xml.dom.minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="  ") 


'''
Language parser for Lua
'''

from .rubylike import RubylikeReader, RubylikeStateMachine


class LuaReader(RubylikeReader):
    # pylint: disable=R0903

    ext = ['lua']
    language_names = ['lua']

    @staticmethod
    def generate_tokens(source_code, addition='', token_class=None):
        return RubylikeReader.generate_tokens(
            source_code,
            r"|\-\-\[\[.*?\]\]" +
            r"|\[\=*\[.*?\]\=*\]" +
            r"|\-\-.*?$" +
            addition)

    def get_comment_from_token(self, token):
        if token.startswith("--"):
            return token

    def __init__(self, context):
        super(LuaReader, self).__init__(context)
        self.parallel_states = [LuaStateMachine(context)]


class LuaStateMachine(RubylikeStateMachine):
    FUNC_KEYWORD = 'function'

    def __init__(self, context):
        super(LuaStateMachine, self).__init__(context)
        self.probable_function_name = None

    def _state_global(self, token):
        if token == "=":
            self.next(self._assigning)
            return
        self.probable_function_name = token
        super(LuaStateMachine, self)._state_global(token)

    def _assigning(self, token):
        if token == self.FUNC_KEYWORD:
            self._state = self._anonymous_def
            return
        self.next(self._state_global, token)

    def _anonymous_def(self, token):
        if token != '(':
            self.next(self._def, token)
            return
        self.context.push_new_function(self.probable_function_name)
        self.next(self._def_parameters)



'''
Get Duplicated parameter lists
'''
from __future__ import print_function
from collections import deque
from itertools import groupby
from .default_ordered_dict import DefaultOrderedDict
from .extension_base import ExtensionBase


class CodeSnippet(object):
    def __init__(self, start_line, end_line, file_name):
        self.start_line = start_line
        self.end_line = end_line
        self.file_name = file_name

    def __str__(self):
        return "%s:%s ~ %s" % (self.file_name, self.start_line, self.end_line)

    def __repr__(self):
        return str(self)

    def fun_yet_to_come(self):
        pass

    def fun_yet_to_come2(self):
        pass


class Sequence(object):
    def __init__(self, start_line):
        self.hash = ''
        self.start_line = self.end_line = start_line

    def append_token(self, unified_token, end_line):
        self.hash += unified_token
        self.end_line = end_line

    def __str__(self):
        return "<%s-%s: %s>" % (self.start_line, self.end_line, self.hash)

    def __repr__(self):
        return str(self)


class InvolvingScope(object):
    def __init__(self, boundaries, keyfunc):
        self.current_file_duplicates = []
        self.dup_starts = set()
        self.boundaries = boundaries
        self.keyfunc = keyfunc

    def same_beginning(self, same, before_same):
        new_starts = [x for x in same if x not in self.dup_starts]
        self.dup_starts |= before_same
        if len(new_starts) > 1:
            for dup in self._duplicate_sequences(new_starts):
                self.current_file_duplicates.append(dup)
                yield dup

    def fun_yet_to_come2(self):
        pass

    def _duplicate_sequences(self, same):
        sequences = [(n, n) for n in same]
        queues = deque([sequences])
        while queues:
            sequences = queues.popleft()
            if self._full_inclusive_sequences(sequences):
                continue
            nexts = [(s, n + 1) for s, n in sequences
                     if n+1 not in self.boundaries
                     and n+1 not in self.dup_starts]
            nexts = sorted(nexts, key=self.keyfunc)
            full_duplicate_stopped = len(nexts) < len(sequences)
            for _, group in groupby(nexts, self.keyfunc):
                group = list(group)
                if len(group) > 1:
                    queues.append(group)
                else:
                    full_duplicate_stopped = True
            if full_duplicate_stopped:
                yield sequences

    def _full_inclusive_sequences(self, sequences):
        return any(
            len(dup) == len(sequences) and
            dup[0][0] <= sequences[0][0] and
            dup[0][1] >= sequences[0][1]
            for dup in self.current_file_duplicates)


class DuplicateFinder(object):
    def __init__(self, nodes, boundaries, **options):
        collapse_repeat_tokens = options.get("collapse_repeat_tokens", 20)
        self.min_duplicate_tokens = options.get("min_duplicate_tokens", 0) * 2
        self.sample_size = options.get("sample_size", 0)
        self.duplicate_token_count = 0
        self.nodes = nodes
        self.boundaries = set(boundaries + [len(nodes)])
        self.hashed_node_indice = DefaultOrderedDict(list)
        for i, node_hash in enumerate(n.hash for n in nodes):
            if i in self.boundaries:
                recent = deque([''] * collapse_repeat_tokens)
            if node_hash not in recent:
                self.hashed_node_indice[node_hash].append(i)
            recent.append(node_hash)
            recent.popleft()

    def find_start_and_ends(self):
        for i, same in enumerate(self.hashed_node_indice.values()):
            if i in self.boundaries:
                scope = InvolvingScope(self.boundaries, self._keyfunc)
            before_same = set(n - self.sample_size for n in same)
            for dup in scope.same_beginning(
                    same, before_same):
                token_count = len(dup) * \
                        (dup[0][1] - dup[0][0] + self.sample_size)
                if token_count >= self.min_duplicate_tokens:
                    self.duplicate_token_count += token_count
                    yield dup

    def duplicate_rate(self):
        try:
            return self.duplicate_token_count * 1.0 / (
                    len(self.nodes) +
                    (len(self.boundaries) - 1) * (self.sample_size - 2))
        except ZeroDivisionError:
            return 0

    def unique_rate(self):
        try:
            return len(self.hashed_node_indice) / len(self.nodes)
        except ZeroDivisionError:
            return 0

    def _keyfunc(self, seq):
        try:
            return self.nodes[seq[1]].hash
        except IndexError:
            return ''


class NestingStackWithUnifiedTokens(object):

    SAMPLE_SIZE = 31
    IGNORE_CONSTANT_VALUE_COUNT = 10

    def __init__(self, decorated):
        self._decorated = decorated
        self.token_register = {}
        self.constant_count = 0
        self.current_scope = set()
        self.scope_stack = [self.current_scope]
        self.unified_tokens = []
        self.previous_token = None

    def __getattr__(self, attr):
        return getattr(self._decorated, attr)

    def add_bare_nesting(self):
        self.current_scope = set()
        self.scope_stack.append(self.current_scope)
        return self._decorated.add_bare_nesting()

    def pop_nesting(self):
        if len(self.scope_stack) > 1:
            for token in self.scope_stack.pop():
                del self.token_register[token]
            self.current_scope = self.scope_stack[-1]
        return self._decorated.pop_nesting()

    @staticmethod
    def _is_const(token):
        return token[0].isdigit() or token[0] in ("'", '"')

    def _unified_token(self, token):
        if self.previous_token in (".", "->"):
            return token
        if token == '-':
            return '+'
        if self._is_const(token):
            return '1'
        if token[0].isalpha():
            if token not in self.token_register:
                self.token_register[token] = 'v'+str(len(self.current_scope))
                self.current_scope.add(token)
            return self.token_register[token]
        return token

    def enqueue_token(self, token, current_line):
        try:
            if self._is_const(self.unified_tokens[-self.SAMPLE_SIZE]):
                self.constant_count -= 1
        except IndexError:
            pass
        if self._is_const(token):
            self.constant_count += 1
        if self.constant_count <= self.SAMPLE_SIZE / 5:
            token = self._unified_token(token)
        self.unified_tokens.append((token, current_line,))
        self.previous_token = token

    def samples(self):
        buf = deque()
        for unified_token, current_line in self.unified_tokens:
            buf.append(Sequence(current_line))
            for code_hash in buf:
                code_hash.append_token(unified_token, current_line)
            if len(buf) > self.SAMPLE_SIZE:
                yield buf.popleft()


class xlizardExtension(ExtensionBase):

    def __init__(self, context=None):
        self.nodes = []
        self.fileinfos = []
        self.saved_duplicate_rate = None
        self.saved_unique_rate = None
        super(xlizardExtension, self).__init__(context)

    def __call__(self, tokens, reader):
        token_unifier = reader.context.decorate_nesting_stack(
                NestingStackWithUnifiedTokens)
        for token in tokens:
            token_unifier.enqueue_token(token, reader.context.current_line)
            yield token
        reader.context.fileinfo.hash_nodes = list(token_unifier.samples())

    def cross_file_process(self, fileinfos):
        for fileinfo in fileinfos:
            self.fileinfos.append((len(self.nodes), fileinfo))
            self.nodes += fileinfo.hash_nodes
            yield fileinfo

    def get_duplicates(self, min_duplicate_tokens=70):
        boundaries = [info[0] for info in self.fileinfos]
        duplicate_finder = DuplicateFinder(
                self.nodes,
                boundaries,
                min_duplicate_tokens=min_duplicate_tokens,
                sample_size=NestingStackWithUnifiedTokens.SAMPLE_SIZE)
        for start_and_ends in duplicate_finder.find_start_and_ends():
            yield self._create_code_snippets(start_and_ends)
        self.saved_duplicate_rate = duplicate_finder.duplicate_rate()
        self.saved_unique_rate = duplicate_finder.unique_rate()

    def duplicate_rate(self):
        return self.saved_duplicate_rate

    def unique_rate(self):
        return self.saved_unique_rate

    def _create_code_snippets(self, start_and_ends):
        return [
            CodeSnippet(
                self.nodes[start].start_line,
                self.nodes[end].end_line,
                self._get_fileinfo_by_token_index(start).filename)
            for start, end in start_and_ends]

    def _get_fileinfo_by_token_index(self, index):
        last_fileinfo = (-1, None)
        for fileinfo in self.fileinfos:
            if fileinfo[0] > index:
                break
            last_fileinfo = fileinfo
        return last_fileinfo[1]

    def print_result(self):
        print("Duplicates")
        print("===================================")
        for dup in self.get_duplicates():
            print("Duplicate block:")
            print("--------------------------")
            for snippet in dup:
                print(snippet)
            print("^^^^^^^^^^^^^^^^^^^^^^^^^^")
            print("")
        print("Total duplicate rate: %.2f%%" % (self.duplicate_rate() * 100))
        print("Total unique rate: %.2f%%" % (self.unique_rate() * 100))



"""Base stuff for extensions"""
from xlizard_languages.code_reader import CodeStateMachine


class ExtensionBase(CodeStateMachine):
    """Base class for all xlizard extensions"""

    def __call__(self, tokens, reader=None):
        self.context = reader.context
        for token in tokens:
            self._state(token)
            yield token



'''
Language parser for Vue.js files
'''

from .code_reader import CodeReader, CodeStateMachine
from .typescript import TypeScriptReader


class VueReader(TypeScriptReader):
    # pylint: disable=R0903

    ext = ['vue']
    language_names = ['vue', 'vuejs']

    def __init__(self, context):
        super(VueReader, self).__init__(context)

    @staticmethod
    def generate_tokens(source_code, addition='', token_class=None):
        # Use the base token generator but ensure we capture Vue block tags
        addition = addition + r"|(?:\<\/?\w+.*?\>)"
        return TypeScriptReader.generate_tokens(source_code, addition, token_class)

    def preprocess(self, tokens):
        current_block = None

        for token in tokens:
            if token.startswith('<script'):
                current_block = 'script'
            elif token.startswith('</script'):
                current_block = None
            elif current_block == 'script':
                if not token.isspace() or token == '\n':
                    yield token



'''
Fan in and Fan out (again)
'''
from collections import Counter
from .extension_base import ExtensionBase


def get_all_indices(value, token_list):
    indices = []
    idx = -1
    while True:
        try:
            idx = token_list.index(value, idx + 1)
            indices.append(idx)
        except ValueError:
            break
    return indices


class xlizardExtension(ExtensionBase):
    '''
        Runs the xlizard Extension for the calculation of structural
        fan-in and fan-out of every procedures in the source code.

        Explanation of terms:
        'ExtensionBase': Base class for all xlizard extensions

        Calculates:
                'fan_in':                     Number of structural fan-in
                'fan_out':		              Number of structural fan-out
                                              in the local scope
                'general_fan_out'             Number of structural fan-out
                                              in the global scope
    '''

    FUNCTION_INFO = {
        "fan_in": {
            "caption": " fan_in ",
            "average_caption": " avg_fan_in "},
        "fan_out": {
            "caption": " fan_out ",
            "average_caption": " avg_fan_out "},
        "general_fan_out": {
            "caption": " general_fan_out ",
            "average_caption": " avg_general_fan_out "}
    }

    def __init__(self, context=None):
        self.all_methods = {}
        super(xlizardExtension, self).__init__(context)

    def _state_global(self, token):
        if not hasattr(self.context.current_function, 'tokens'):
            self.context.current_function.tokens = list()
        self.context.current_function.tokens.append(token)

    def cross_file_process(self, fileinfos):
        for fileinfo in fileinfos:
            try:
                new_funcs = {f.unqualified_name: f for f in
                             fileinfo.function_list}
                self.all_methods.update(new_funcs)
                self._add_to_fan_outs(new_funcs.keys())
                self._add_to_general_fan_out()
                self._add_to_fan_ins(fileinfo.function_list)
            except (AttributeError, TypeError, ValueError):
                pass
            yield fileinfo

    def _add_to_fan_outs(self, keys):
        for other_func in self.all_methods.values():
            intersect = Counter(keys) & Counter(other_func.tokens)
            other_func.fan_out += len(list(intersect.elements()))

    def _add_to_general_fan_out(self):
        structures = set(['if', 'else', 'elif', 'for', 'foreach', 'while',
                          'do', 'try', 'catch', 'switch', 'finally',
                          'except', 'with'])
        punctuations = set(['(', ')', '{', '}'])
        for other_func in self.all_methods.values():
            bracket_indexes = get_all_indices('(', other_func.tokens)
            for idx in bracket_indexes[1:]:
                # Suggestion ?
                if other_func.tokens[idx - 1] not in (structures |
                                                      punctuations):
                    other_func.general_fan_out += 1

    def _add_to_fan_ins(self, new_funcs):
        for func in new_funcs:
            for name in func.tokens:
                if name in self.all_methods:
                    self.all_methods[name].fan_in += 1



'''
Language parser for Go lang
'''

from .code_reader import CodeReader
from .clike import CCppCommentsMixin
from .golike import GoLikeStates


class GoReader(CodeReader, CCppCommentsMixin):
    # pylint: disable=R0903

    ext = ['go']
    language_names = ['go']

    def __init__(self, context):
        super(GoReader, self).__init__(context)
        self.parallel_states = [GoStates(context)]

    @staticmethod
    def generate_tokens(source_code, addition='', token_class=None):
        addition = addition + r"|`[^`]*`"  # Add support for backtick-quoted strings
        return CodeReader.generate_tokens(source_code, addition, token_class)

    def __call__(self, tokens, reader):
        self.context = reader.context
        for token in tokens:
            # Skip counting ? in backtick-quoted strings
            if token.startswith('`') and token.endswith('`'):
                for state in self.parallel_states:
                    state(token)
                yield token
                continue

            # For non-backtick tokens, process normally
            for state in self.parallel_states:
                state(token)
            yield token
        for state in self.parallel_states:
            state.statemachine_before_return()
        self.eof()


class GoStates(GoLikeStates):  # pylint: disable=R0903
    pass



#!/usr/bin/env python
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#
"""
xlizard is an extensible Cyclomatic Complexity Analyzer for many programming
languages including C/C++ (doesn't require all the header files).
"""
from __future__ import print_function, division
import codecs
import sys
import itertools
import re
import os
from fnmatch import fnmatch
import hashlib
from . import version 

if sys.version[0] == '2':
    from future_builtins import map, filter  # pylint: disable=W0622, F0401

try:
    from .xlizard_languages import languages, get_reader_for, CLikeReader
except ImportError as e:
    sys.stderr.write(f"Import error details: {str(e)}\n")
    sys.exit(2)
try:
    from .xlizard_ext import version
    from .xlizard_ext import print_xml
    from .xlizard_ext import print_csv
    from .xlizard_ext import html_output
    from .xlizard_ext import auto_open, auto_read
    from .xlizard_ext import print_checkstyle
except ImportError:
    sys.stderr.write("Cannot find the xlizard_ext modules.")

DEFAULT_CCN_THRESHOLD, DEFAULT_WHITELIST, \
    DEFAULT_MAX_FUNC_LENGTH = 15, "whitexlizard.txt", 1000


# pylint: disable-msg=too-many-arguments
def analyze(paths, exclude_pattern=None, threads=1, exts=None,
            lans=None):
    '''
    returns an iterator of file information that contains function
    statistics.
    '''
    exclude_pattern = exclude_pattern or []
    files = get_all_source_files(paths, exclude_pattern, lans)
    return analyze_files(files, threads, exts)


def analyze_files(files, threads=1, exts=None):
    extensions = exts or get_extensions([])
    file_analyzer = FileAnalyzer(extensions)
    result = map_files_to_analyzer(files, file_analyzer, threads)
    for extension in extensions:
        if hasattr(extension, 'cross_file_process'):
            result = extension.cross_file_process(result)
    return result


def _extension_arg(parser):
    parser.add_argument("-E", "--extension",
                        help='''User the extensions. The available extensions
                        are: -Ecpre: it will ignore code in the #else branch.
                        -Ewordcount: count word frequencies and generate tag
                        cloud. -Eoutside: include the global code as one
                        function.  -EIgnoreAssert: to ignore all code in
                        assert. -ENS: count nested control structures.''',
                        action="append",
                        dest="extensions",
                        default=[])


def arg_parser(prog=None):
    from argparse import ArgumentParser, Action, ArgumentError

    class DictAction(Action):  # pylint: disable=R0903
        def __init__(self, option_strings, dest, nargs=None, **kwargs):
            super(DictAction, self).__init__(option_strings, dest, **kwargs)

        def __call__(self, parser, namespace, value, option_string=None):
            if not re.match(r"\s*\w+\s*=\s*\d+", value):
                raise ArgumentError(self, "should be like nloc=20")
            k, val = value.split("=", 2)
            getattr(namespace, self.dest)[k.strip()] = int(val.strip())

    parser = ArgumentParser(prog=prog)
    parser.add_argument('paths', nargs='*', default=['.'],
                        help='list of the filename/paths.')
    parser.add_argument('--version', action='version', version=version)
    parser.add_argument("-l", "--languages",
                        help='''List the programming languages you want to
                        analyze. if left empty, it'll search for all languages
                        it knows. `xlizard -l cpp -l java`searches for C++ and
                        Java code. The available languages are:
    ''' + ', '.join(x.language_names[0] for x in languages()),
                        action="append",
                        dest="languages",
                        default=[])
    parser.add_argument("-V", "--verbose",
                        help="Output in verbose mode (long function name)",
                        action="store_true",
                        dest="verbose",
                        default=False)
    parser.add_argument("-C", "--CCN",
                        help='''Threshold for cyclomatic complexity number
                        warning. The default value is %d.
                        Functions with CCN bigger than it will generate warning
                        ''' % DEFAULT_CCN_THRESHOLD,
                        type=int,
                        dest="CCN",
                        default=DEFAULT_CCN_THRESHOLD)
    parser.add_argument("-f", "--input_file",
                        help='''get a list of filenames from the given file
                        ''',
                        type=str,
                        dest="input_file")
    parser.add_argument("-o", "--output_file",
                        help='''Output file. The output format is inferred
                        from the file extension (e.g. .html), unless it is
                        explicitly specified (e.g. using --xml).
                        ''',
                        type=str,
                        dest="output_file")
    parser.add_argument("-L", "--length",
                        help='''Threshold for maximum function length
                        warning. The default value is %d.
                        Functions length bigger than it will generate warning
                        ''' % DEFAULT_MAX_FUNC_LENGTH,
                        type=int,
                        dest="length",
                        default=DEFAULT_MAX_FUNC_LENGTH)
    parser.add_argument("-a", "--arguments",
                        help="Limit for number of parameters",
                        type=int, dest="arguments", default=100)
    parser.add_argument("-w", "--warnings_only",
                        help='''Show warnings only, using clang/gcc's warning
                        format for printing warnings.
                        http://clang.llvm.org/docs/UsersManual.html#cmdoption-fdiagnostics-format
                        ''',
                        action="store_const",
                        const=print_clang_style_warning,
                        dest="printer")
    parser.add_argument("--warning-msvs",
                        help='''Show warnings only, using Visual Studio's
                        warning format for printing warnings.
                        https://msdn.microsoft.com/en-us/library/yxkt8b26.aspx
                        ''',
                        action="store_const",
                        const=print_msvs_style_warning,
                        dest="printer")
    parser.add_argument("-i", "--ignore_warnings",
                        help='''If the number of warnings is equal or less
                        than the number, the tool will exit normally;
                        otherwise, it will generate error.
                        If the number is negative,
                        the tool exits normally
                        regardless of the number of warnings.
                        Useful in makefile for legacy code.''',
                        type=int,
                        dest="number",
                        default=0)
    parser.add_argument("-x", "--exclude",
                        help='''Exclude files that match the pattern. * matches
                        everything,
                        ? matches any single character, "./folder/*" exclude
                        everything in the folder recursively. Multiple patterns
                        can be specified. Don't forget to add "" around the
                        pattern.''',
                        action="append",
                        dest="exclude",
                        default=[])
    parser.add_argument("-t", "--working_threads",
                        help='''number of working threads. The default
                        value is 1. Using a bigger
                        number can fully utilize the CPU and often faster.''',
                        type=int,
                        dest="working_threads",
                        default=1)
    parser.add_argument("-X", "--xml",
                        help='''Generate XML in cppncss style instead of the
                        tabular output. Useful to generate report in Jenkins
                        server''',
                        action="store_const",
                        const=print_xml,
                        dest="printer")
    parser.add_argument("--csv",
                        help='''Generate CSV output as a transform of the
                        default output''',
                        action="store_const",
                        const=print_csv,
                        dest="printer")
    parser.add_argument("-H", "--html",
                        help='''Output HTML report''',
                        action="store_const",
                        const=html_output,
                        dest="printer")
    parser.add_argument("-m", "--modified",
                        help='''Calculate modified cyclomatic complexity number
                        , which count a switch/case with multiple cases as
                        one CCN.''',
                        action="append_const",
                        const="modified",
                        dest="extensions",
                        default=[])
    parser.add_argument("--checkstyle",
                        help='''Generate Checkstyle XML output for integration with Jenkins and other tools.''',
                        action="store_const",
                        const=print_checkstyle,
                        dest="printer")
    _extension_arg(parser)
    parser.add_argument("-s", "--sort",
                        help='''Sort the warning with field. The field can be
                        nloc, cyclomatic_complexity, token_count,
                        parameter_count, etc. Or an customized field.''',
                        action="append",
                        dest="sorting",
                        default=[])
    parser.add_argument("-T", "--Threshold",
                        help='''Set the limit for a field. The field can be
                        nloc, cyclomatic_complexity, token_count,
                        parameter_count, etc. Or an customized file. xlizard
                        will report warning if a function exceed the limit''',
                        action=DictAction,
                        dest="thresholds",
                        default={})
    parser.add_argument("-W", "--whitelist",
                        help='''The path and file name to the whitelist file.
                        It's './whitexlizard.txt' by default.
                        Find more information in README.''',
                        type=str,
                        dest="whitelist",
                        default=DEFAULT_WHITELIST)

    parser.usage = '''xlizard [options] [PATH or FILE] [PATH] ...'''
    parser.description = __doc__
    return parser


class Nesting(object):  # pylint: disable=R0903
    '''
    Nesting represent one level of nesting in any programming language.
    '''
    @property
    def name_in_space(self):
        return ''


BARE_NESTING = Nesting()


class Namespace(Nesting):  # pylint: disable=R0903

    def __init__(self, name):
        self.name = name

    @property
    def name_in_space(self):
        return self.name + "::" if self.name else ''


class FunctionInfo(Nesting):  # pylint: disable=R0902

    def __init__(self, name, filename, start_line=0, ccn=1):
        self.cyclomatic_complexity = ccn
        self.nloc = 1
        self.token_count = 1  # the first token
        self.name = name
        self.long_name = name
        self.start_line = start_line
        self.end_line = start_line
        self.full_parameters = []
        self.filename = filename
        self.top_nesting_level = -1
        self.fan_in = 0
        self.fan_out = 0
        self.general_fan_out = 0
        self.max_nesting_depth = 0  # Initialize max_nesting_depth to 0

    @property
    def name_in_space(self):
        return self.name + "."

    @property
    def unqualified_name(self):
        '''
        name without qualification like namespaces or classes.
        Just the bare name without '::'.
        '''
        return self.name.split('::')[-1]

    location = property(lambda self:
                        f" {self.name}@{self.start_line}-{self.end_line}@{self.filename}")

    parameter_count = property(lambda self: len(self.parameters))

    @property
    def parameters(self):
        # Exclude empty tokens as parameters. These can occur in languages
        # allowing a trailing comma on the last parameter in an function
        # argument list.
        # Regex matches the parameter name, then optionally:
        # - a default value given after an '=' sign
        # - a type annotation given after a ':'
        matches = [re.search(r'(\w+)(\s=.*)?(\s:.*)?$', f)
                   for f in self.full_parameters]
        return [m.group(1) for m in matches if m]

    @property
    def length(self):
        return self.end_line - self.start_line + 1

    def add_to_function_name(self, app):
        self.name += app
        self.long_name += app

    def add_to_long_name(self, app):
        if self.long_name:
            if self.long_name[-1].isalpha() and app[0].isalpha():
                self.long_name += ' '
        self.long_name += app

    def add_parameter(self, token):
        self.add_to_long_name(" " + token)

        if not self.full_parameters:
            self.full_parameters.append(token)
        elif token == ",":
            self.full_parameters.append('')
        else:
            self.full_parameters[-1] += " " + token


class FileInformation(object):  # pylint: disable=R0903

    def __init__(self, filename, nloc, function_list=None):
        self.filename = filename
        self.nloc = nloc
        self.function_list = function_list or []
        self.token_count = 0

    average_nloc = property(lambda self: self.functions_average("nloc"))
    average_token_count = property(
        lambda self: self.functions_average("token_count"))
    average_cyclomatic_complexity = property(
        lambda self: self.functions_average("cyclomatic_complexity"))
    CCN = property(
        lambda self:
        sum(fun.cyclomatic_complexity for fun in self.function_list))
    ND = property(  # pylint: disable=C0103
        lambda self:
        sum(fun.max_nesting_depth for fun in self.function_list))

    def functions_average(self, att):
        summary = sum(getattr(fun, att) for fun in self.function_list)
        return summary / len(self.function_list) if self.function_list else 0


class NestingStack(object):

    def __init__(self):
        self.nesting_stack = []
        self.pending_function = None

    def with_namespace(self, name):
        return ''.join([x.name_in_space for x in self.nesting_stack] + [name])

    def add_bare_nesting(self):
        self.nesting_stack.append(self._create_nesting())

    def add_namespace(self, token):
        self.pending_function = None
        self.nesting_stack.append(Namespace(token))

    def start_new_function_nesting(self, function):
        self.pending_function = function

    def _create_nesting(self):
        tmp = self.pending_function
        self.pending_function = None
        if tmp:
            return tmp
        return BARE_NESTING

    def pop_nesting(self):
        self.pending_function = None
        if self.nesting_stack:
            return self.nesting_stack.pop()

    @property
    def current_nesting_level(self):
        return len(self.nesting_stack)

    @property
    def last_function(self):
        funs = [f for f in self.nesting_stack if isinstance(f, FunctionInfo)]
        return funs[-1] if funs else None


class FileInfoBuilder(object):
    '''
    The builder is also referred as "context" in the code,
    because each language readers use this builder to build
    source file and function information and the builder keep
    the context information that's needed for the building.
    '''

    def __init__(self, filename):
        self.fileinfo = FileInformation(filename, 0)
        self.current_line = 0
        self.forgive = False
        self.forgive_global = False
        self.newline = True
        self.global_pseudo_function = FunctionInfo('*global*', filename, 0)
        self.current_function = self.global_pseudo_function
        self.stacked_functions = []
        self._nesting_stack = NestingStack()

    def __getattr__(self, attr):
        # delegating to _nesting_stack
        return getattr(self._nesting_stack, attr)

    def decorate_nesting_stack(self, decorate_class):
        self._nesting_stack = decorate_class(self._nesting_stack)
        return self._nesting_stack

    def pop_nesting(self):
        nest = self._nesting_stack.pop_nesting()
        if isinstance(nest, FunctionInfo):
            endline = self.current_function.end_line
            self.end_of_function()
            self.current_function = (
                    self._nesting_stack.last_function or
                    self.global_pseudo_function)
            self.current_function.end_line = endline

    def add_nloc(self, count):
        self.fileinfo.nloc += count
        self.current_function.nloc += count
        self.current_function.end_line = self.current_line
        self.newline = count > 0

    def try_new_function(self, name):
        self.current_function = FunctionInfo(
            self.with_namespace(name),
            self.fileinfo.filename,
            self.current_line)
        self.current_function.top_nesting_level = self.current_nesting_level

    def confirm_new_function(self):
        self.start_new_function_nesting(self.current_function)
        self.current_function.cyclomatic_complexity = 1

    def restart_new_function(self, name):
        self.try_new_function(name)
        self.confirm_new_function()

    def push_new_function(self, name):
        self.stacked_functions.append(self.current_function)
        self.restart_new_function(name)

    def add_condition(self, inc=1):
        self.current_function.cyclomatic_complexity += inc

    def add_to_long_function_name(self, app):
        self.current_function.add_to_long_name(app)

    def add_to_function_name(self, app):
        self.current_function.add_to_function_name(app)

    def parameter(self, token):
        self.current_function.add_parameter(token)

    def end_of_function(self):
        if not self.forgive:
            if self.current_function.name != '*global*' or not self.forgive_global:
                self.fileinfo.function_list.append(self.current_function)
        self.forgive = False
        if self.stacked_functions:
            self.current_function = self.stacked_functions.pop()
        else:
            self.current_function = self.global_pseudo_function


def preprocessing(tokens, reader):
    if hasattr(reader, "preprocess"):
        return reader.preprocess(tokens)
    return (t for t in tokens if not t.isspace() or t == '\n')


def comment_counter(tokens, reader):
    for token in tokens:
        comment = reader.get_comment_from_token(token)
        if comment is not None:
            for _ in comment.splitlines()[1:]:
                yield '\n'
            if comment.strip().startswith("#xlizard forgive global"):
                reader.context.forgive_global = True
            elif comment.strip().startswith("#xlizard forgive"):
                reader.context.forgive = True
            if "GENERATED CODE" in comment:
                return
        else:
            yield token


def line_counter(tokens, reader):
    context = reader.context
    context.current_line = 1
    newline = 1
    for token in tokens:
        if token != "\n":
            count = token.count('\n')
            context.current_line += count
            context.add_nloc(count + newline)
            newline = 0
            yield token
        else:
            context.current_line += 1
            newline = 1


def token_counter(tokens, reader):
    context = reader.context
    for token in tokens:
        context.fileinfo.token_count += 1
        context.current_function.token_count += 1
        yield token


def condition_counter(tokens, reader):
    conditions = reader.conditions
    for token in tokens:
        if token in conditions:
            reader.context.add_condition()
        yield token


class FileAnalyzer(object):  # pylint: disable=R0903

    def __init__(self, extensions):
        self.processors = extensions

    def __call__(self, filename):
        try:
            return self.analyze_source_code(
                filename, auto_read(filename))
        except UnicodeDecodeError:
            sys.stderr.write("Error: doesn't support none utf encoding '%s'\n"
                             % filename)
        except IOError:
            sys.stderr.write("Error: Fail to read source file '%s'\n"
                             % filename)
        except IndexError:
            sys.stderr.write("Error: Fail to parse file '%s'\n"
                             % filename)
            raise
        return FileInformation(filename, 0, [])

    def analyze_source_code(self, filename, code):
        context = FileInfoBuilder(filename)
        reader = (get_reader_for(filename) or CLikeReader)(context)
        tokens = reader.generate_tokens(code)
        try:
            for processor in self.processors:
                tokens = processor(tokens, reader)
            for _ in reader(tokens, reader):
                pass
        except RecursionError as e:
            sys.stderr.write(
                "[skip] fail to process '%s' with RecursionError - %s\n" %
                (filename, e))
        return context.fileinfo


def map_files_to_analyzer(files, analyzer, working_threads):
    mapmethod = get_map_method(working_threads)
    return mapmethod(analyzer, files)


def warning_filter(option, module_infos):
    for file_info in module_infos:
        if file_info:
            for fun in file_info.function_list:
                if any(getattr(fun, attr) > limit for attr, limit in
                       option.thresholds.items()):
                    yield fun


def whitelist_filter(warnings, script=None, whitelist=None):
    def _get_whitelist_item(script):
        white = {}
        pieces = script.replace('::', '##').split(':')
        if len(pieces) > 1:
            white['file_name'] = pieces[0]
            script = pieces[1]
        white['function_names'] = (
            [x.strip().replace('##', '::') for x in script.split(',')])
        return white

    def _in_list(warning):
        return any(_match_whitelist_item(white, warning)
                   for white in whitelist)

    def _match_whitelist_item(white, warning):
        return (warning.name in white['function_names'] and
                warning.filename == white.get('file_name', warning.filename))

    def get_whitelist(whitelist):
        if os.path.isfile(whitelist):
            return open(whitelist, mode='r').read()
        if whitelist != DEFAULT_WHITELIST:
            print("WARNING: !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            print("WARNING: the whitelist \""+whitelist+"\" doesn't exist.")
        return ''

    if not script:
        script = get_whitelist(whitelist)
    whitelist = [
        _get_whitelist_item(line.split('#')[0])
        for line in script.splitlines()]
    for warning in warnings:
        if not _in_list(warning):
            yield warning


class OutputScheme(object):
    '''
    Collect the schema of the data columns.
    Each extension can define some additional data columns to
    the FunctionInfo structure, or even add properties to
    the FileInformation structure.

    In any extension class, define a class level variable:

        FUNCTION_INFO = {
            'column_name' : {
                'caption': 'if defined, will show the column in result',
                'average_caption': 'if defined, will add average function
                                    to FileInformation and show in the
                                    end result.
            }
        }
    '''

    def __init__(self, ext):
        self.extensions = ext
        self.items = [
            {
                'caption': "  NLOC  ", 'value': "nloc",
                'avg_caption': ' Avg.NLOC '},
            {
                'caption': "  CCN  ", 'value': "cyclomatic_complexity",
                'avg_caption': ' AvgCCN '},
            {
                'caption': " token ", 'value': "token_count",
                'avg_caption': ' Avg.token '},
            {'caption': " PARAM ", 'value': "parameter_count"},
            {'caption': " length ", 'value': "length"},
            ] + [
            {
                'caption': caption,
                'value': part,
                'avg_caption': average
            }
            for caption, part, average in self._ext_member_info()]
        self.items.append({'caption': " location  ", 'value': 'location'})

    def patch_for_extensions(self):
        def _patch(name):
            setattr(FileInformation, "average_" + name,
                    property(lambda self: self.functions_average(name)))
        for item in self.items:
            if 'avg_caption' in item:
                _patch(item["value"])

    def any_silent(self):
        return any(hasattr(ex, 'silent_all_others') for ex in self.extensions)

    def value_columns(self):
        return [item['value'] for item in self.items]

    def _ext_member_info(self):
        for ext in self.extensions:
            if hasattr(ext, "FUNCTION_INFO"):
                for key in ext.FUNCTION_INFO:
                    yield (
                        ext.FUNCTION_INFO[key].get("caption", None),
                        key,
                        ext.FUNCTION_INFO[key].get("average_caption", None))

    def captions(self):
        caps = [item.get('caption') for item in self.items]
        return "".join(caption for caption in caps if caption)

    @staticmethod
    def _head(captions):
        return "\n".join(("=" * len(captions), captions, "-" * len(captions)))

    def function_info_head(self):
        return self._head(self.captions())

    def function_info(self, fun):
        return ''.join(
            str(getattr(fun, item['value'])).rjust(len(item['caption']))
            for item in self.items if item['caption'])

    def average_captions(self):
        return "".join([
            e['avg_caption'] for e in self.items
            if e.get("avg_caption", None)])

    def average_formatter(self):
        return "".join([
            "{{module.average_{ext[value]}:{size}.1f}}"
            .format(ext=e, size=len(e['avg_caption']))
            for e in self.items
            if e.get("avg_caption", None)])

    def clang_warning_format(self):
        return ("{f.filename}:{f.start_line}: warning: {f.name} has {f.nloc} NLOC, "
                "{f.cyclomatic_complexity} CCN, {f.token_count} token, {f.parameter_count} PARAM, "
                "{f.length} length, {f.max_nesting_depth} ND")

    def msvs_warning_format(self):
        return (
            "{f.filename}({f.start_line}): warning: {f.name} ({f.long_name}) has " +
            ", ".join([
                "{{f.{ext[value]}}} {caption}"
                .format(ext=e, caption=e['caption'].strip())
                for e in self.items[:-1]
                ]))


def print_warnings(option, scheme, warnings):
    warning_count = 0
    warning_nloc = 0
    warn_str = "!!!! Warnings ({0}) !!!!".format(
        ' or '.join("{0} > {1}".format(
            k, val) for k, val in option.thresholds.items()))
    for warning in warnings:
        if warning_count == 0:
            print("\n" + "=" * len(warn_str) + "\n" + warn_str)
            print(scheme.function_info_head())
        warning_count += 1
        warning_nloc += warning.nloc
        print(scheme.function_info(warning))
    if warning_count == 0:
        print_no_warnings(option)
    return warning_count, warning_nloc


def print_no_warnings(option):
    warn_str = "No thresholds exceeded ({0})".format(
        ' or '.join("{0} > {1}".format(
            k, val) for k, val in option.thresholds.items()))
    print("\n" + "=" * len(warn_str) + "\n" + warn_str)


class AllResult(object):
    def __init__(self, result):
        self.result = list(file_info for file_info in result if file_info)
        self.all_fun = list(itertools.chain(*(file_info.function_list
                                            for file_info in self.result)))

    def function_count(self):
        return len(self.all_fun) or 1

    def nloc_in_functions(self):
        return sum([f.nloc for f in self.all_fun]) or 1

    def as_fileinfo(self):
        return FileInformation(
                    "",
                    sum([f.nloc for f in self.result]),
                    self.all_fun)


def print_total(warning_count, warning_nloc, all_result, scheme):
    print("=" * 90)
    print("Total nloc  " + scheme.average_captions() + "  Fun Cnt  Warning"
          " cnt   Fun Rt   nloc Rt")
    print("-" * 90)
    print((
        "{module.nloc:10d}" +
        scheme.average_formatter() +
        "{function_count:9d}{warning_count:13d}" +
        "{function_rate:10.2f}{nloc_rate:8.2f}").format(
                  module=all_result.as_fileinfo(),
                  function_count=all_result.function_count(),
                  warning_count=warning_count,
                  function_rate=(warning_count/all_result.function_count()),
                  nloc_rate=(warning_nloc/all_result.nloc_in_functions())))


def print_and_save_modules(all_fileinfos, scheme):
    saved_fileinfos = []
    print(scheme.function_info_head())
    for module_info in all_fileinfos:
        if module_info:
            saved_fileinfos.append(module_info)
            for fun in module_info.function_list:
                try:
                    print(scheme.function_info(fun))
                except UnicodeEncodeError:
                    print("Found ill-formatted unicode function name.")
    print("%d file analyzed." % (len(saved_fileinfos)))
    print("==============================================================")
    print("NLOC   " + scheme.average_captions() + " function_cnt    file")
    print("--------------------------------------------------------------")
    for module_info in saved_fileinfos:
        print((
            "{module.nloc:7d}" +
            scheme.average_formatter() +
            "{function_count:10d}" +
            "     {module.filename}").format(
            module=module_info,
            function_count=len(module_info.function_list)))
    return saved_fileinfos


def get_warnings(code_infos, option):
    warnings = whitelist_filter(warning_filter(option, code_infos),
                                whitelist=option.whitelist)
    if isinstance(option.sorting, list) and option.sorting:
        warnings = sorted(warnings, reverse=True, key=lambda x: getattr(
            x, option.sorting[0]))
    return warnings


def print_result(result, option, scheme, total_factory):
    result = print_and_save_modules(result, scheme)
    warnings = get_warnings(result, option)
    warning_count, warning_nloc = print_warnings(option, scheme, warnings)
    print_total(warning_count, warning_nloc, total_factory(result), scheme)
    return warning_count


def silent_printer(result, *_):
    '''
    just to exhaust the result, no output.
    '''
    for _ in result:
        pass
    return 0


def print_clang_style_warning(code_infos, option, scheme, _):
    count = 0
    for warning in get_warnings(code_infos, option):
        print(scheme.clang_warning_format().format(f=warning))
        count += 1
    return count


def print_msvs_style_warning(code_infos, option, scheme, _):
    count = 0
    for warning in get_warnings(code_infos, option):
        print(scheme.msvs_warning_format().format(f=warning))
        count += 1
    return count


def get_map_method(working_threads):
    try:
        if working_threads == 1:
            raise ImportError
        import multiprocessing
        pool = multiprocessing.Pool(processes=working_threads)
        return pool.imap_unordered
    except ImportError:
        return map


def md5_hash_file(full_path_name):
    ''' return md5 hash of a file '''
    try:
        with auto_open(full_path_name, mode='r') as source_file:
            if sys.version_info[0] == 3:
                code_md5 = hashlib.md5(source_file.read().encode('utf-8'))
            else:
                code_md5 = hashlib.md5(source_file.read())
        return code_md5.hexdigest()
    except IOError:
        return None
    except UnicodeDecodeError:
        return None
    except UnicodeEncodeError:
        return None


def get_all_source_files(paths, exclude_patterns, lans):
    '''
    Function counts md5 hash for the given file and checks if it isn't a
    duplicate using set of hashes for previous files.

    If a .gitignore file is found in any of the given paths, it will be used
    to filter out files that match the gitignore patterns.
    '''
    hash_set = set()
    gitignore_spec = None
    base_path = None

    def _load_gitignore():
        nonlocal gitignore_spec, base_path
        try:
            import pathspec
            for path in paths:
                gitignore_path = os.path.join(path, '.gitignore')
                if os.path.exists(gitignore_path):
                    with open(gitignore_path, 'r') as gitignore_file:
                        # Read lines and strip whitespace and empty lines
                        patterns = [line.strip() for line in gitignore_file.readlines()]
                        patterns = [p for p in patterns if p and not p.startswith('#')]
                        gitignore_spec = pathspec.PathSpec.from_lines('gitwildmatch', patterns)
                        base_path = path
                    break
        except ImportError:
            pass

    def _support(reader):
        return not lans or set(lans).intersection(
            reader.language_names)

    def _validate_file(pathname):
        if gitignore_spec is not None and base_path is not None:
            rel_path = os.path.relpath(pathname, base_path)
            # Normalize path separators for consistent matching
            rel_path = rel_path.replace(os.sep, '/')
            if gitignore_spec.match_file(rel_path):
                return False
        return (
            pathname in paths or (
                get_reader_for(pathname) and
                _support(get_reader_for(pathname)) and
                all(not fnmatch(pathname, p) for p in exclude_patterns) and
                _not_duplicate(pathname)))

    def _not_duplicate(full_path_name):
        fhash = md5_hash_file(full_path_name)
        if not fhash or fhash not in hash_set:
            hash_set.add(fhash)
            return True

    def all_listed_files(paths):
        for path in paths:
            if os.path.isfile(path):
                yield path
            else:
                for root, _, files in os.walk(path, topdown=False):
                    for filename in files:
                        yield os.path.join(root, filename)

    _load_gitignore()
    return filter(_validate_file, all_listed_files(paths))


def parse_args(argv):
    def extend_parser(parser_to_extend):
        from argparse import ArgumentParser
        parser = ArgumentParser(add_help=False)
        _extension_arg(parser)
        opt, _ = parser.parse_known_args(args=argv[1:])
        extensions = get_extensions(opt.extensions)
        for ext in extensions:
            if hasattr(ext, "set_args"):
                ext.set_args(parser_to_extend)  # pylint: disable=E1101
        return parser_to_extend
    parser = extend_parser(arg_parser(argv[0]))
    opt = parser.parse_args(args=argv[1:])
    opt.extensions = get_extensions(opt.extensions)
    values = OutputScheme(opt.extensions).value_columns()
    no_fields = (set(opt.sorting) | set(opt.thresholds.keys())) - set(values)
    if no_fields:
        error_message = "Wrong field name '%s'.\n" % ", ".join(no_fields)
        error_message += "Candidates are: " + ', '.join(values) + "\n"
        sys.stderr.write(error_message)
        sys.exit(2)
    if "cyclomatic_complexity" not in opt.thresholds:
        opt.thresholds["cyclomatic_complexity"] = opt.CCN
    if "max_nesting_depth" not in opt.thresholds and hasattr(opt, "ND"):
        opt.thresholds["max_nesting_depth"] = opt.ND
    if "max_nested_structures" not in opt.thresholds and hasattr(opt, "NS"):
        opt.thresholds["max_nested_structures"] = opt.NS
    if "length" not in opt.thresholds:
        opt.thresholds["length"] = opt.length
    if "nloc" not in opt.thresholds:
        opt.thresholds["nloc"] = 1000000
    if "parameter_count" not in opt.thresholds:
        opt.thresholds["parameter_count"] = 5
        opt.thresholds.update({
        'comment_percentage': 20,
        'max_block_depth': 5,
        'pointer_operations': 20,
        'preprocessor_directives': 15
    })
    if opt.output_file:
        inferred_printer = infer_printer_from_file_ext(opt.output_file)
        if inferred_printer:
            # Always use print_checkstyle for .checkstyle.xml
            if opt.output_file.lower().endswith('.checkstyle.xml'):
                opt.printer = inferred_printer
            elif not opt.printer:
                opt.printer = inferred_printer
            elif opt.printer != inferred_printer:
                msg = "Warning: overriding output file extension.\n"
                sys.stderr.write(msg)
    return opt


def infer_printer_from_file_ext(path):
    lower_path = path.lower()
    if lower_path.endswith(".checkstyle.xml"):
        return print_checkstyle
    if lower_path.endswith(".html"):
        return html_output
    if lower_path.endswith(".xml"):
        return print_xml
    if lower_path.endswith(".csv"):
        return print_csv
    return None


def open_output_file(path):
    try:
        return codecs.open(path, 'w', encoding='utf8')
    except OSError:
        msg = "Error: failed to open output file '{}'\n.".format(path)
        sys.stderr.write(msg)
        sys.exit(2)


def get_extensions(extension_names):
    from importlib import import_module as im

    def expand_extensions(existing):
        for name in extension_names:
            ext = (
                    im('xlizard_ext.xlizard' + name.lower())
                    .xlizardExtension()
                    if isinstance(name, str) else name)
            existing.insert(
                len(existing) if not hasattr(ext, "ordering_index") else
                ext.ordering_index,
                ext)
        return existing

    return expand_extensions([
            preprocessing,
            comment_counter,
            line_counter,
            token_counter,
            condition_counter,
        ])


analyze_file = FileAnalyzer(get_extensions([]))  # pylint: disable=C0103


def main(argv=None):
    """Command-line entrance to xlizard.

    Args:
        argv: Arguments vector; if None, sys.argv by default.
    """
    options = parse_args(argv or sys.argv)
    printer = options.printer or print_result
    schema = OutputScheme(options.extensions)
    if schema.any_silent():
        printer = silent_printer
    schema.patch_for_extensions()
    if options.input_file:
        options.paths = auto_read(options.input_file).splitlines()
    original_stdout = sys.stdout
    output_file = None
    result = analyze(
        options.paths,
        options.exclude,
        options.working_threads,
        options.extensions,
        options.languages)
    warning_count = None
    if options.output_file:
        output_file = open_output_file(options.output_file)
        sys.stdout = output_file
        # Special handling for checkstyle output
        if getattr(printer, "__name__", "") == "print_checkstyle":
            warning_count = printer(result, options, schema, AllResult, file=output_file)
        else:
            warning_count = printer(result, options, schema, AllResult)
    else:
        # Special handling for checkstyle output
        if getattr(printer, "__name__", "") == "print_checkstyle":
            warning_count = printer(result, options, schema, AllResult, file=original_stdout)
        else:
            warning_count = printer(result, options, schema, AllResult)
    print_extension_results(options.extensions)
    list(result)
    if output_file:
        sys.stdout = original_stdout
        output_file.close()
    if 0 <= options.number < warning_count:
        sys.exit(1)


def print_extension_results(extensions):
    for extension in extensions:
        if hasattr(extension, 'print_result'):
            extension.print_result()


if __name__ == "__main__":
    main()



'''
This is an extension of xlizard,
It lets xlizard to use modified cyclomatic complexity number,
where the whole switch/case will be counted as 1.
'''


class xlizardExtension(object):  # pylint: disable=R0903

    def __call__(self, tokens, reader):
        for token in tokens:
            if token == 'switch':
                reader.context.add_condition()
                if hasattr(reader.context, "add_nd_condition"):
                    reader.context.add_nd_condition()
            elif token == 'case':
                reader.context.add_condition(-1)
                if hasattr(reader.context, "add_nd_condition"):
                    reader.context.add_nd_condition(-1)
            yield token



"""
This extension dumps all the comments in one file
"""
from __future__ import print_function, division


DEFAULT_NS_THRESHOLD = 3


class xlizardExtension(object):  # pylint: disable=R0903

    silent_all_others = True
    ordering_index = 1

    def __call__(self, tokens, reader):
        first = True
        print("=" * 20)
        print(reader.context.fileinfo.filename)
        print("-" * 20)
        for token in tokens:
            comment = reader.get_comment_from_token(token)
            if comment is not None:
                if not first or "copyright" not in comment.lower():
                    print(comment)
                first = False
            yield token



'''
generate token with javascript style regular expression.
'''

import re


def js_style_regex_expression(func):
    def generate_tokens_with_regex(source_code, addition='', token_class=None):
        regx_regx = r"\/(\S*?[^\s\\]\/)+?(igm)*"
        regx_pattern = re.compile(regx_regx)
        tokens = list(func(source_code, addition, token_class))
        result = []
        i = 0
        while i < len(tokens):
            token = tokens[i]
            if token == '/':
                # Check if this could be a regex pattern
                is_regex = False
                if i == 0:
                    is_regex = True
                elif i > 0:
                    prev_token = tokens[i-1].strip()
                    if prev_token and prev_token[-1] in '=,({[?:!&|;':
                        is_regex = True

                if is_regex:
                    # This is likely a regex pattern start
                    regex_tokens = [token]
                    i += 1
                    while i < len(tokens) and not tokens[i].endswith('/'):
                        regex_tokens.append(tokens[i])
                        i += 1
                    if i < len(tokens):
                        regex_tokens.append(tokens[i])
                        i += 1
                        # Check for regex flags
                        if i < len(tokens) and re.match(r'^[igm]+$', tokens[i]):
                            regex_tokens.append(tokens[i])
                            i += 1
                    combined = ''.join(regex_tokens)
                    if regx_pattern.match(combined):
                        result.append(combined)
                    else:
                        result.extend(regex_tokens)
                else:
                    # This is a division operator
                    result.append(token)
            else:
                result.append(token)
            i += 1
        return result
    return generate_tokens_with_regex



'''
This is an extension of xlizard,
It helps to deal with C code with preprocessors that
is hard to parse. It works by always ignoring the code
between #else and #end.
'''


class xlizardExtension(object):  # pylint: disable=R0903

    ordering_index = 0

    def __call__(self, tokens, reader):
        def preprocess_tokens(tokens):
            else_count = 0
            if_stack = []
            for token in tokens:
                if token.startswith("#"):
                    if_stack.append(token)
                    else_count += token.count("#else")
                    if token.startswith("#endif"):
                        while if_stack:
                            last = if_stack.pop()
                            else_count -= last.count("#else")
                            if last.startswith("#if"):
                                break
                    for _ in range(token.count('\n')):
                        yield '\n'
                elif else_count:
                    for _ in range(token.count('\n')):
                        yield '\n'
                elif not (if_stack and if_stack[-1].startswith("#elif")):
                    yield token

        if "c" not in reader.ext:
            return tokens
        return preprocess_tokens(tokens)



"""
This extension counts nested control structures within a function.

The extension is implemented with C++ and Python in mind,
but it is expected to work with other languages supported by xlizard
with its language reader implementing 'nesting_level' metric for tokens.

The code borrows heavily from implementation of Nesting Depth extension
originally written by Mehrdad Meh and Terry Yin.
"""

from xlizard import FunctionInfo
from xlizard_ext.xlizardnd import patch_append_method
from xlizard_languages.code_reader import CodeStateMachine
from .extension_base import ExtensionBase


DEFAULT_NS_THRESHOLD = 3


class xlizardExtension(ExtensionBase):  # pylint: disable=R0903
    """The intent of the code is to detect control structures as entities.

    The implementation relies on nesting level metric for tokens
    provided by language readers.
    If the following contract for the nesting level metric does not hold,
    this implementation of nested structure counting is invalid.

    If a control structure has started its block (e.g., '{'),
    and its level is **less** than the next structure,
    the next structure is nested.

    If a control structure has *not* started its block,
    and its level is **no more** than the next structure,
    the next structure is nested (compound statement).

    If a control structure level is **higher** than the next structure,
    it is considered closed.

    If a control structure has started its block,
    and its level is **equal** to the next structure,
    it is considered closed.

    The level of any non-structure tokens is treated
    with the same logic as for the next structures
    for control block **starting** and **closing** purposes.
    """

    FUNCTION_INFO = {"max_nested_structures": {"caption": "  NS  "}}

    # TODO: Delegate this to language readers  # pylint: disable=fixme
    structures = set(['if', 'else', 'elif', 'for', 'foreach', 'while',
                      'do', 'try', 'catch', 'switch', 'finally',
                      'except', 'with'])
    matching_structures = set(['else', 'elif', 'catch', 'finally'])

    @staticmethod
    def set_args(parser):
        parser.add_argument(
            "--NS",
            help='''Threshold for the number of nested control structures.
            The default value is %d.
            ''' % DEFAULT_NS_THRESHOLD,
            type=int,
            dest="NS",
            default=DEFAULT_NS_THRESHOLD)

    def __init__(self):
        super(xlizardExtension, self).__init__(None)
        self.structure_piles = [0]

    def pile_up_within_block(self):
        self.structure_piles[-1] += 1
        cur_level = sum(self.structure_piles)
        # Is there a path around _state_global?
        if not hasattr(self.context.current_function, "max_nested_structures"):
            self.context.current_function.max_nested_structures = 0
        if self.context.current_function.max_nested_structures < cur_level:
            self.context.current_function.max_nested_structures = cur_level

    def _state_global(self, token):
        # Necessary to prevent functions without nested blocks to fail
        if not hasattr(self.context.current_function, "max_nested_structures"):
            self.context.current_function.max_nested_structures = 0
        if token == '{':
            self.structure_piles.append(0)
        elif token in ';}':
            if token == '}':
                self.structure_piles.pop()
            self._state = self._block_ending
        elif token in self.structures:
            self._state = self._in_structure_head

    @CodeStateMachine.read_inside_brackets_then("()")
    def _in_structure_head(self, token):
        self.pile_up_within_block()
        self._state = self._state_global
        self._state(token)

    def _block_ending(self, token):
        if token in self.matching_structures:
            self.structure_piles[-1] -= 1
        else:
            self.structure_piles[-1] = 0
        self._state = self._state_global
        self._state(token)


def _init_nested_structure_data(self, *_):
    self.max_nested_structures = 0


patch_append_method(_init_nested_structure_data, FunctionInfo, "__init__")



'''
Language parser for Apple Swift
'''

from .code_reader import CodeReader, CodeStateMachine
from .clike import CCppCommentsMixin
from .golike import GoLikeStates


class SwiftReplaceLabel:
    def preprocess(self, tokens):
        tokens = list(t for t in tokens if not t.isspace() or t == '\n')

        def replace_label(tokens, target, replace):
            for i in range(0, len(tokens) - len(target)):
                if tokens[i:i + len(target)] == target:
                    for j, repl in enumerate(replace):
                        tokens[i + j] = repl
            return tokens

        for k in (k for k in self.conditions if k.isalpha()):
            tokens = replace_label(tokens, ["(", k, ":"], ["(", "_" + k, ":"])
            tokens = replace_label(tokens, [",", k, ":"], [",", "_" + k, ":"])
        return tokens


class SwiftReader(CodeReader, CCppCommentsMixin, SwiftReplaceLabel):
    # pylint: disable=R0903

    FUNC_KEYWORD = 'def'
    ext = ['swift']
    language_names = ['swift']
    _conditions = set(['if', 'for', 'while', '&&', '||', '?', 'catch',
                      'case', 'guard'])

    def __init__(self, context):
        super(SwiftReader, self).__init__(context)
        self.parallel_states = [SwiftStates(context)]

    @staticmethod
    def generate_tokens(source_code, addition='', token_class=None):
        return CodeReader.generate_tokens(
            source_code,
            r"|`\w+`" +
            r"|\w+\?" +
            r"|\w+\!" +
            r"|\?\?" +
            addition)


class SwiftStates(GoLikeStates):  # pylint: disable=R0903
    def _state_global(self, token):
        if token in ('init', 'subscript'):
            self.context.push_new_function('')
            self.next(self._function_name, token)
        elif token in ('get', 'set', 'willSet', 'didSet', 'deinit'):
            self.context.push_new_function(token)
            self._state = self._expect_function_impl
        elif token == 'protocol':
            self._state = self._protocol
        elif token in ('let', 'var', 'case', ','):
            self._state = self._expect_declaration_name
        else:
            super(SwiftStates, self)._state_global(token)

    def _expect_declaration_name(self, token):
        self._state = self._state_global

    @CodeStateMachine.read_inside_brackets_then("{}")
    def _protocol(self, end_token):
        if end_token == "}":
            self._state = self._state_global



"""
This is an extension of xlizard, that counts the 'Nesting Depth'
in every function.
"""
from xlizard import FileInfoBuilder, FunctionInfo

DEFAULT_ND_THRESHOLD = 7


class xlizardExtension(object):  # pylint: disable=R0903

    FUNCTION_INFO = {
            "max_nesting_depth": {
                "caption": "  ND  ",
                "average_caption": " Avg.ND "}}

    @staticmethod
    def set_args(parser):
        parser.add_argument(
            "-N", "--ND",
            help='''Threshold for nesting depth number
            warning. The default value is %d.
            Functions with ND bigger than it will generate warning
            ''' % DEFAULT_ND_THRESHOLD,
            type=int,
            dest="ND",
            default=DEFAULT_ND_THRESHOLD)

    def __call__(self, tokens, reader, l_depth=0):  # pylint: disable=R0912
        if hasattr(reader, "loops"):
            loops = reader.loops
        else:
            loops = set(['if', 'foreach', 'for', 'while', '&&', '||',
                         '?', 'catch', 'case', 'try', 'def'])
        if hasattr(reader, "bracket"):
            bracket = reader.bracket
        else:
            bracket = '}'
        if hasattr(reader, "loop_indicator"):
            loop_indicator = reader.loop_indicator
        else:
            loop_indicator = '{'
        if hasattr(reader, "indent_indicator"):
            indent_indicator = reader.indent_indicator
        else:
            indent_indicator = ';'
        for token in tokens:
            if token in loops:
                l_depth = reader.context.add_nd_condition()
                if not reader.context.get_loop_status():
                    reader.context.add_hidden_bracket_condition()
                    reader.context.loop_bracket_status()
            if token == loop_indicator:
                reader.context.loop_bracket_status()
            if token == bracket:
                l_depth = reader.context.add_nd_condition(-1)
            if token == indent_indicator:
                hidden_brackets = reader.context.get_hidden_bracket()
                check_loop_brackets(reader, l_depth, hidden_brackets)
            if l_depth < 0:
                l_depth = 0
                reader.context.reset_nd_complexity()
            yield token


def check_loop_brackets(reader, l_depth, hidden_brackets):
    if hidden_brackets > 0:
        reader.context.add_hidden_bracket_condition(-1)
        l_depth = reader.context.add_nd_condition(-1)


class NDFileInfoAddition(FileInfoBuilder):

    def add_nd_condition(self, inc=1):
        self.current_function.nesting_depth += inc
        nd_tmp = self.current_function.nesting_depth
        if self.current_function.max_nesting_depth < nd_tmp:
            self.current_function.max_nesting_depth = nd_tmp
        return self.current_function.nesting_depth

    def reset_nd_complexity(self):
        self.current_function.nesting_depth = 0
        self.current_function.hidden_bracket = 0
        self.current_function.bracket_loop = False

    def add_hidden_bracket_condition(self, inc=1):
        self.current_function.hidden_bracket += inc

    def get_hidden_bracket(self):
        return self.current_function.hidden_bracket

    def loop_bracket_status(self):
        tmp_bracket_loop = self.current_function.bracket_loop
        self.current_function.bracket_loop = not tmp_bracket_loop

    def get_loop_status(self):
        return self.current_function.bracket_loop


def get_method(cls, name):
    """ python3 doesn't need the __func__ to get the func of the
        method.
    """
    method = getattr(cls, name)
    if hasattr(method, "__func__"):
        method = method.__func__
    return method


def patch(frm, accept_class):
    for method in [k for k in frm.__dict__ if not k.startswith("_")]:
        setattr(accept_class, method, get_method(frm, method))


def patch_append_method(frm, accept_class, method_name):
    old_method = get_method(accept_class, method_name)

    def appended(*args, **kargs):
        old_method(*args, **kargs)
        frm(*args, **kargs)

    setattr(accept_class, method_name, appended)


def _init_nesting_depth_data(self, *_):
    self.nesting_depth = 0
    self.max_nesting_depth = 0
    self.hidden_bracket = 0
    self.bracket_loop = False


patch(NDFileInfoAddition, FileInfoBuilder)
patch_append_method(_init_nesting_depth_data, FunctionInfo, "__init__")



'''
DefaultOrderedDict is a combinated OrderedDict and
defaultdict.
'''
from collections import OrderedDict


class DefaultOrderedDict(OrderedDict):
    def __init__(self, default_factory=None):
        OrderedDict.__init__(self)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = (self.default_factory,)
        return type(self), args, None, None, self.items()



''' programming languages of xlizard '''

from .zig import ZigReader
from .clike import CLikeReader
from .erlang import ErlangReader
from .java import JavaReader
from .javascript import JavaScriptReader
from .kotlin import KotlinReader
from .python import PythonReader
from .objc import ObjCReader
from .ttcn import TTCNReader
from .swift import SwiftReader
from .ruby import RubyReader
from .csharp import CSharpReader
from .php import PHPReader
from .scala import ScalaReader
from .gdscript import GDScriptReader
from .go import GoReader
from .lua import LuaReader
from .rust import RustReader
from .typescript import TypeScriptReader
from .fortran import FortranReader
from .solidity import SolidityReader
from .jsx import JSXReader
from .tsx import TSXReader
from .vue import VueReader
from .perl import PerlReader

__all__ = ['languages', 'get_reader_for']


def languages():
    return [
        CLikeReader,
        JavaReader,
        CSharpReader,
        JavaScriptReader,
        PythonReader,
        ObjCReader,
        TTCNReader,
        RubyReader,
        PHPReader,
        SwiftReader,
        ScalaReader,
        GDScriptReader,
        GoReader,
        LuaReader,
        RustReader,
        TypeScriptReader,
        FortranReader,
        KotlinReader,
        SolidityReader,
        ErlangReader,
        ZigReader,
        JSXReader,
        TSXReader,
        VueReader,
        PerlReader,
    ]


def get_reader_for(filename):
    """Находит подходящий Reader для файла"""
    for lan in languages():
        if lan.match_filename(filename):
            return lan


'''
Language parser for Scala
'''
from .code_reader import CodeReader
from .clike import CCppCommentsMixin
from .golike import GoLikeStates
__author__ = 'David Baum'


class ScalaReader(CodeReader, CCppCommentsMixin):
    # pylint: disable=R0903

    ext = ['scala']
    language_names = ['scala']
    _conditions = set(['if', 'for', 'while', '&&', '||', '?', 'catch',
                      'case', 'do'])

    def __init__(self, context):
        super(ScalaReader, self).__init__(context)
        self.parallel_states = [ScalaStates(context)]


class ScalaStates(GoLikeStates):  # pylint: disable=R0903

    FUNC_KEYWORD = 'def'

    def _state_global(self, token):
        super(ScalaStates, self)._state_global(token)

    def _expect_function_impl(self, token):
        if token == "=":
            self._state = self._expect_function_body
        else:
            super(ScalaStates, self)._expect_function_impl(token)

    def _expect_function_body(self, token):
        if self.context.newline:
            self.context.end_of_function()
            self.next(self._state_global, token)
        elif token == '{':
            self.sub_state(ScalaStates(self.context))

    def statemachine_before_return(self):
        if self._state == self._expect_function_body:
            self.context.end_of_function()



"""
Language parser for erlang
"""

import re
from .code_reader import CodeReader, CodeStateMachine
import pygments.token as py_token
from pygments import lex, lexers


class ErlangReader(CodeReader):
    # pylint: disable=R0903

    ext = ['erl', 'hrl', 'es', 'escript']
    language_names = ['erlang']
    _conditions = {'and', 'case', 'catch', 'if', 'not', 'or', '?', 'when'}

    def __init__(self, context):
        super(ErlangReader, self).__init__(context)
        self.parallel_states = [ErlangStates(context)]

    @staticmethod
    def get_comment_from_token(token):
        if token.startswith("%%"):
            return token[2:]

    @staticmethod
    def generate_tokens(source_code, addition='', token_class=None):
        lexer = lexers.get_lexer_by_name('erlang')
        tokens = lex(source_code, lexer=lexer)
        return map(
            lambda x: x[1],
            filter(lambda x: x[0] != py_token.Whitespace, tokens)
        )


class ErlangStates(CodeStateMachine):
    # pylint: disable=R0903

    func_name_pattern = re.compile("[a-zA-Z]+[a-zA-Z0-9_]*")

    def __init__(self, context):
        super(ErlangStates, self).__init__(context)
        self.punctuated = False

    def _state_global(self, token):
        if token == '-':
            self.punctuated = True
        elif self.func_name_pattern.match(token) and not self.punctuated:
            self.context.push_new_function(token)
            self._state = self._state_after_name
        elif token == 'end':
            self._state = self._state_nested_end
        elif token == '.' or token == ';':
            self.statemachine_return()
        else:
            self.punctuated = False

    def _state_after_name(self, token):
        if token == '(':
            self._state = self._state_start_of_params
            self.context.add_to_long_function_name(token)
            self.lbr = 1
            self.rbr = 0
        else:
            self.func_match_failed(token)

    def _state_start_of_params(self, token):
        if token == ')':
            self.rbr += 1
            if self.lbr == self.rbr:
                self._state = self._state_end_of_params
                self.context.add_to_long_function_name(" " + token)
                self.punctuated = False
                return
        if token == '(':
            self.lbr += 1
        self.context.parameter(token)

    def _state_end_of_params(self, token):
        if token == '-':
            self.punctuated = True
        elif token == '>' and self.punctuated:
            if (len(self.context.stacked_functions) <= 1 or
                    self.context.current_function.name == 'fun'):
                self.next(self._state_func_first_line, token)
        else:
            self.func_match_failed(token)

    def _state_func_first_line(self, _):
        def callback():
            self._state = self._state_global
            self.context.end_of_function()

        self.sub_state(self.statemachine_clone(), callback)
        self.punctuated = False

    def _state_nested_end(self, token):
        if token == '.' or token == ',':
            if (len(self.context.stacked_functions) > 1 and
                    self.context.stacked_functions[-1].name == 'fun'):
                self.statemachine_return()
                return

        self._state = self._state_global

    def func_match_failed(self, token):
        self.punctuated = False
        self._state = self._state_global
        curr_cyc_comp = self.context.current_function.cyclomatic_complexity - 1
        if self.context.stacked_functions:
            self.context.current_function = self.context.stacked_functions.pop()
        else:
            self.context.current_function = self.context.global_pseudo_function
        self.context.add_condition(curr_cyc_comp)
        self.next(self._state_global, token)



'''
This is an extension of xlizard, that counts the amount of goto's
'''


class xlizardExtension():  # pylint: disable=R0903

    FUNCTION_INFO = {"goto_count": {"caption": " goto's "}}

    def __call__(self, tokens, reader):
        for token in tokens:
            if not hasattr(reader.context.current_function, "goto_count"):
                reader.context.current_function.goto_count = 0
            if token == "goto":
                reader.context.current_function.goto_count += 1
            yield token



# -*- coding: utf-8 -*-
"""
Module for combining xlizard and SourceMonitor metrics
"""
from typing import Dict, List, Optional

class CombinedMetrics:
    def __init__(self, xlizard_data, sm_metrics: Optional[Dict] = None):
        self.xlizard = xlizard_data
        self.sourcemonitor = sm_metrics or {
            'comment_percentage': 0,
            'max_block_depth': 0,
            'pointer_operations': 0,
            'preprocessor_directives': 0
        }
        
    @property
    def filename(self):
        return self.xlizard.filename
        
    @property
    def functions(self):
        return self.xlizard.function_list
        
    @property
    def basename(self):
        return self.xlizard.filename.split('/')[-1]
        
    @property
    def dirname(self):
        dirs = self.xlizard.filename.split('/')[:-1]
        return '/'.join(dirs) if dirs else "Project Files"


'''
Language parser for JavaScript
'''

from .typescript import TypeScriptReader


class JavaScriptReader(TypeScriptReader):
    # pylint: disable=R0903

    ext = ['js', 'cjs', 'mjs']
    language_names = ['javascript', 'js']

    def __init__(self, context):
        super(JavaScriptReader, self).__init__(context)
